@@ -31,7 +31,7 @@ char offset0[4] = {0x00, 0x00, 0x00, 0x00};
  * portability, payload types must be defined with their fields in the right
  * order.
  */
-#if defined(_ISOC99_SOURCE)
+#if defined(_ISOC99_SOURCE) || defined(__clang__)
 // ISO C99's tagged syntax
 #define TYPE(val)		.type=(val)
 #define CLOCK_RATE(val)		.clock_rate=(val)
@@ -597,7 +597,7 @@ PayloadType payload_type_silk_nb={
 	BITS_PER_SAMPLE( 0),
 	ZERO_PATTERN(NULL),
 	PATTERN_LENGTH( 0),
-	NORMAL_BITRATE(10000),
+	NORMAL_BITRATE(13000),
 	MIME_TYPE ("SILK"),
 	CHANNELS(1)
 };
@@ -667,3 +667,14 @@ PayloadType payload_type_opus = {
 	MIME_TYPE("opus"),
 	CHANNELS(1)
 };
+
+PayloadType payload_type_isac = {
+	TYPE( PAYLOAD_AUDIO_PACKETIZED),
+	CLOCK_RATE(16000),
+	BITS_PER_SAMPLE(0),
+	ZERO_PATTERN(NULL),
+	PATTERN_LENGTH(0),
+	NORMAL_BITRATE(32000),
+	MIME_TYPE("iSAC"),
+	CHANNELS(1)
+};
