@@ -20,7 +20,7 @@
 
 #if defined(WIN32) || defined(_WIN32_WCE)
 #include "ortp-config-win32.h"
-#else
+#elif HAVE_CONFIG_H
 #include "ortp-config.h"
 #endif
 
@@ -32,6 +32,14 @@
 #include "utils.h"
 #include "rtpsession_priv.h"
 
+#if (_WIN32_WINNT >= 0x0600)
+#include <delayimp.h>
+#undef ExternC /* avoid redefinition... */
+#ifndef WINAPI_FAMILY_PHONE_APP
+#include <QOS2.h>
+#endif
+#endif
+
 extern mblk_t *rtcp_create_simple_bye_packet(uint32_t ssrc, const char *reason);
 extern int rtcp_sr_init(RtpSession *session, char *buf, int size);
 extern int rtcp_rr_init(RtpSession *session, char *buf, int size);
@@ -41,8 +49,8 @@ extern int rtcp_rr_init(RtpSession *session, char *buf, int size);
 /* this function initialize all session parameter's that depend on the payload type */
 static void payload_type_changed(RtpSession *session, PayloadType *pt){
 	jitter_control_set_payload(&session->rtp.jittctl,pt);
-	session->rtp.rtcp_report_snt_interval=RTCP_DEFAULT_REPORT_INTERVAL*pt->clock_rate;
 	rtp_session_set_time_jump_limit(session,session->rtp.time_jump);
+	rtp_session_set_rtcp_report_interval(session,session->rtcp.interval);
 	if (pt->type==PAYLOAD_VIDEO){
 		session->permissive=TRUE;
 		ortp_message("Using permissive algorithm");
@@ -90,7 +98,11 @@ extern void rtp_parse(RtpSession *session, mblk_t *mp, uint32_t local_str_ts,
 
 
 static uint32_t uint32_t_random(){
+#ifdef HAVE_ARC4RANDOM
+	return arc4random();
+#else
 	return random();
+#endif
 }
 
 
@@ -213,8 +225,13 @@ mblk_t *rtp_getq_permissive(queue_t *q,uint32_t timestamp, int *rejected)
 void
 rtp_session_init (RtpSession * session, int mode)
 {
+	JBParameters jbp;
+	if (session == NULL) 
+	{
+	    ortp_debug("rtp_session_init: Invalid paramter (session=NULL)");
+	    return;
+	}
 	memset (session, 0, sizeof (RtpSession));
-	session->rtp.max_rq_size = 100;/* maximum number of packet allowed to be queued */
 	session->mode = (RtpSessionMode) mode;
 	if ((mode == RTP_SESSION_RECVONLY) || (mode == RTP_SESSION_SENDRECV))
 	{
@@ -228,13 +245,20 @@ rtp_session_init (RtpSession * session, int mode)
 		session->snd.ssrc=uint32_t_random();
 		/* set default source description */
 		rtp_session_set_source_description(session,"unknown@unknown",NULL,NULL,
-				NULL,NULL,"oRTP-" ORTP_VERSION,"This is free sofware (LGPL) !");
+				NULL,NULL,"oRTP-" ORTP_VERSION,NULL);
 	}
 	session->snd.telephone_events_pt=-1;	/* not defined a priori */
 	session->rcv.telephone_events_pt=-1;	/* not defined a priori */
 	rtp_session_set_profile (session, &av_profile); /*the default profile to work with */
 	session->rtp.socket=-1;
 	session->rtcp.socket=-1;
+#ifndef WIN32
+	session->rtp.snd_socket_size=0;	/*use OS default value unless on windows where they are definitely too short*/
+	session->rtp.rcv_socket_size=0;
+#else
+	session->rtp.snd_socket_size=session->rtp.rcv_socket_size=65536;
+#endif
+	session->rtp.ssrc_changed_thres=50;
 	session->dscp=RTP_DEFAULT_DSCP;
 	session->multicast_ttl=RTP_DEFAULT_MULTICAST_TTL;
 	session->multicast_loopback=RTP_DEFAULT_MULTICAST_LOOPBACK;
@@ -256,12 +280,22 @@ rtp_session_init (RtpSession * session, int mode)
 	rtp_session_set_send_payload_type(session,0);
 	/*sets supposed recv payload type to undefined */
 	rtp_session_set_recv_payload_type(session,-1);
-	rtp_session_set_jitter_compensation(session,RTP_DEFAULT_JITTER_TIME);
-	rtp_session_enable_adaptive_jitter_compensation(session,FALSE);
+	/* configure jitter buffer with working default parameters */
+	jbp.min_size=RTP_DEFAULT_JITTER_TIME;
+	jbp.nom_size=RTP_DEFAULT_JITTER_TIME;
+	jbp.max_size=-1;
+	jbp.max_packets= 100;/* maximum number of packet allowed to be queued */
+	jbp.adaptive=TRUE;
+	rtp_session_enable_jitter_buffer(session,TRUE);
+	rtp_session_set_jitter_buffer_params(session,&jbp);
 	rtp_session_set_time_jump_limit(session,5000);
+	rtp_session_enable_rtcp(session,TRUE);
+	rtp_session_set_rtcp_report_interval(session,RTCP_DEFAULT_REPORT_INTERVAL);
 	session->recv_buf_size = UDP_MAX_SIZE;
 	session->symmetric_rtp = FALSE;
 	session->permissive=FALSE;
+	session->reuseaddr=TRUE;
+	msgb_allocator_init(&session->allocator);
 }
 
 
@@ -279,7 +313,12 @@ rtp_session_new (int mode)
 {
 	RtpSession *session;
 	session = (RtpSession *) ortp_malloc (sizeof (RtpSession));
-	rtp_session_init (session, mode);
+	if (session == NULL)
+	{
+	    ortp_error("rtp_session_new: Memory allocation failed");
+	    return NULL;
+	}
+      rtp_session_init (session, mode);
 	return session;
 }
 
@@ -358,6 +397,39 @@ rtp_session_set_profile (RtpSession * session, RtpProfile * profile)
 	rtp_session_telephone_events_supported(session);
 }
 
+/**
+ *	By default oRTP automatically sends RTCP SR or RR packets. If
+ *	yesno is set to FALSE, the RTCP sending of packet is disabled.
+ *	This functionnality might be needed for some equipments that do not
+ *	support RTCP, leading to a traffic of ICMP errors on the network.
+ *	It can also be used to save bandwidth despite the RTCP bandwidth is 
+ *	actually and usually very very low.
+**/
+void rtp_session_enable_rtcp(RtpSession *session, bool_t yesno){
+	session->rtcp.enabled=yesno;
+}
+
+/**
+ * Sets the default interval in milliseconds for RTCP reports emitted by the session
+ *
+**/
+void rtp_session_set_rtcp_report_interval(RtpSession *session, int value_ms){
+	int recvpt=rtp_session_get_recv_payload_type(session);
+	int sendpt=rtp_session_get_send_payload_type(session);
+	if (recvpt!=-1){
+		PayloadType *pt=rtp_profile_get_payload(session->rcv.profile,recvpt);
+		if (pt!=NULL){
+			session->rtcp.rtcp_report_snt_interval_r=(value_ms*pt->clock_rate)/1000;
+		}
+	}
+	if (sendpt!=-1){
+		PayloadType *pt=rtp_profile_get_payload(session->snd.profile,sendpt);
+		if (pt!=NULL){
+			session->rtcp.rtcp_report_snt_interval_s=(value_ms*pt->clock_rate)/1000;
+		}
+	}
+	session->rtcp.interval=value_ms;
+}
 
 /**
  *	Set the RTP profile to be used for the sending by this session. By default, all session are created by
@@ -426,9 +498,9 @@ RtpProfile *rtp_session_get_recv_profile(RtpSession *session){
 }
 
 /**
- *	The default value is 65535 bytes, a big value which is working for everyone.
- *	However if your application can make assumption on the MTU, it can be interesting
- *	to set it to a lower value in order to save memory.
+ *	The default value is UDP_MAX_SIZE bytes, a value which is working for mostly everyone.
+ *	However if your application can make assumption on the sizes of received packet,
+ *	it can be interesting to set it to a lower value in order to save memory.
  *
  * @param session a rtp session
  * @param bufsize max size in bytes for receiving packets
@@ -437,6 +509,22 @@ void rtp_session_set_recv_buf_size(RtpSession *session, int bufsize){
 	session->recv_buf_size=bufsize;
 }
 
+/**
+ *	Set kernel send maximum buffer size for the rtp socket.
+ *	A value of zero defaults to the operating system default.
+**/
+void rtp_session_set_rtp_socket_send_buffer_size(RtpSession * session, unsigned int size){
+	session->rtp.snd_socket_size=size;
+}
+
+/**
+ *	Set kernel recv maximum buffer size for the rtp socket.
+ *	A value of zero defaults to the operating system default.
+**/
+void rtp_session_set_rtp_socket_recv_buffer_size(RtpSession * session, unsigned int size){
+	session->rtp.rcv_socket_size=size;
+}
+
 /**
  *	This function provides the way for an application to be informed of various events that
  *	may occur during a rtp session. @signal is a string identifying the event, and @cb is 
@@ -511,7 +599,7 @@ rtp_session_signal_disconnect_by_callback (RtpSession * session, const char *sig
 
 
 /**
- * sets the initial sequence number of a sending session.
+ * Set the initial sequence number for outgoing stream..
  * @param session		a rtp session freshly created.
  * @param addr			a 16 bit unsigned number.
  *
@@ -521,10 +609,20 @@ void rtp_session_set_seq_number(RtpSession *session, uint16_t seq){
 }
 
 
+/**
+ * Get the current sequence number for outgoing stream.
+**/
 uint16_t rtp_session_get_seq_number(RtpSession *session){
 	return session->rtp.snd_seq;
 }
 
+/**
+ * Returns the highest extended sequence number received.
+**/
+uint32_t rtp_session_get_rcv_ext_seq_number(RtpSession *session){
+	return session->rtp.hwrcv_extseq;
+}
+
 
 /**
  *	Sets the SSRC for the outgoing stream.
@@ -539,12 +637,32 @@ rtp_session_set_ssrc (RtpSession * session, uint32_t ssrc)
 	session->snd.ssrc = ssrc;
 }
 
+/**
+ *	Get the SSRC for the outgoing stream.
+ *
+ * @param session a rtp session.
+**/
+uint32_t
+rtp_session_get_send_ssrc (RtpSession* session)
+{
+	return session->snd.ssrc;
+}
+
+/**
+ * Get the SSRC for the incoming stream.
+ * 
+ * If no packets have been received yet, 0 is returned.
+**/
+uint32_t rtp_session_get_recv_ssrc(RtpSession *session){
+	return session->rcv.ssrc;
+}
+
 
 void rtp_session_update_payload_type(RtpSession *session, int paytype){
 	/* check if we support this payload type */
 	PayloadType *pt=rtp_profile_get_payload(session->rcv.profile,paytype);
-	session->hw_recv_pt=paytype;
 	if (pt!=0){
+		session->hw_recv_pt=paytype;
 		ortp_message ("payload type changed to %i(%s) !",
 				 paytype,pt->mime_type);
 		payload_type_changed(session,pt);
@@ -644,8 +762,8 @@ static void rtp_header_init_from_session(rtp_header_t *rtp, RtpSession *session)
 
 /**
  *	Allocates a new rtp packet. In the header, ssrc and payload_type according to the session's
- *	context. Timestamp and seq number are not set, there will be set when the packet is going to be
- *	sent with rtp_session_sendm_with_ts().
+ *	context. Timestamp is not set, it will be set when the packet is going to be
+ *	sent with rtp_session_sendm_with_ts(). Sequence number is initalized to previous sequence number sent + 1
  *	If payload_size is zero, thus an empty packet (just a RTP header) is returned.
  *
  *@param session a rtp session.
@@ -732,19 +850,9 @@ mblk_t * rtp_session_create_packet_in_place(RtpSession *session,uint8_t *buffer,
 	return mp;
 }
 
-/**
- *	Send the rtp datagram @mp to the destination set by rtp_session_set_remote_addr() 
- *	with timestamp @timestamp. For audio data, the timestamp is the number
- *	of the first sample resulting of the data transmitted. See rfc1889 for details.
- *  The packet (@mp) is freed once it is sended.
- *
- *@param session a rtp session.
- *@param mp a rtp packet presented as a mblk_t.
- *@param timestamp the timestamp of the data to be sent.
- * @return the number of bytes sent over the network.
-**/
+
 int
-rtp_session_sendm_with_ts (RtpSession * session, mblk_t *mp, uint32_t timestamp)
+__rtp_session_sendm_with_ts (RtpSession * session, mblk_t *mp, uint32_t packet_ts, uint32_t send_ts)
 {
 	rtp_header_t *rtp;
 	uint32_t packet_time;
@@ -755,13 +863,13 @@ rtp_session_sendm_with_ts (RtpSession * session, mblk_t *mp, uint32_t timestamp)
 
 	if (session->flags & RTP_SESSION_SEND_NOT_STARTED)
 	{
-		session->rtp.snd_ts_offset = timestamp;
-        /* Set initial last_rcv_time to first send time. */
-        if ((session->flags & RTP_SESSION_RECV_NOT_STARTED)
-            || session->mode == RTP_SESSION_SENDONLY)
-        {
-            gettimeofday(&session->last_recv_time, NULL);
-        }
+		session->rtp.snd_ts_offset = send_ts;
+		/* Set initial last_rcv_time to first send time. */
+		if ((session->flags & RTP_SESSION_RECV_NOT_STARTED)
+		|| session->mode == RTP_SESSION_SENDONLY)
+		{
+		ortp_gettimeofday(&session->last_recv_time, NULL);
+		}
 		if (session->flags & RTP_SESSION_SCHEDULED)
 		{
 			session->rtp.snd_time_offset = sched->time_;
@@ -774,13 +882,13 @@ rtp_session_sendm_with_ts (RtpSession * session, mblk_t *mp, uint32_t timestamp)
 	 * not block */
 	if (session->flags & RTP_SESSION_SCHEDULED)
 	{
+		wait_point_lock(&session->snd.wp);
 		packet_time =
 			rtp_session_ts_to_time (session,
-				     timestamp -
+				     send_ts -
 				     session->rtp.snd_ts_offset) +
 					session->rtp.snd_time_offset;
 		/*ortp_message("rtp_session_send_with_ts: packet_time=%i time=%i",packet_time,sched->time_);*/
-		wait_point_lock(&session->snd.wp);
 		if (TIME_IS_STRICTLY_NEWER_THAN (packet_time, sched->time_))
 		{
 			wait_point_wakeup_at(&session->snd.wp,packet_time,(session->flags & RTP_SESSION_BLOCKING_MODE)!=0);	
@@ -790,26 +898,35 @@ rtp_session_sendm_with_ts (RtpSession * session, mblk_t *mp, uint32_t timestamp)
 		wait_point_unlock(&session->snd.wp);
 	}
 	
+	if(mp==NULL) {/*for people who just want to be blocked but
+		 do not want to send anything.*/
+		session->rtp.snd_last_ts = packet_ts;
+		return 0;
+	}
 
 	rtp=(rtp_header_t*)mp->b_rptr;
 	
 	packsize = msgdsize(mp) ;
-	
-	rtp->timestamp=timestamp;
-	if (session->snd.telephone_events_pt==rtp->paytype)
-	{
-		session->rtp.snd_seq++;
-		rtp->seq_number = session->rtp.snd_seq;
+
+	if (rtp->version == 0) {
+		/* We are probably trying to send a STUN packet so don't change its content. */
+	} else {
+		rtp->timestamp=packet_ts;
+		if (session->snd.telephone_events_pt==rtp->paytype)
+		{
+			rtp->seq_number = session->rtp.snd_seq;
+			session->rtp.snd_seq++;
+		}
+		else
+			session->rtp.snd_seq=rtp->seq_number+1;
+		session->rtp.snd_last_ts = packet_ts;
+
+		ortp_global_stats.sent += packsize;
+		stream->sent_payload_bytes+=packsize-RTP_FIXED_HEADER_SIZE;
+		stream->stats.sent += packsize;
+		ortp_global_stats.packet_sent++;
+		stream->stats.packet_sent++;
 	}
-	else
-		session->rtp.snd_seq=rtp->seq_number+1;
-	session->rtp.snd_last_ts = timestamp;
-
-
-	ortp_global_stats.sent += packsize;
-	stream->stats.sent += packsize;
-	ortp_global_stats.packet_sent++;
-	stream->stats.packet_sent++;
 
 	error = rtp_session_rtp_send (session, mp);
 	/*send RTCP packet if needed */
@@ -820,6 +937,24 @@ rtp_session_sendm_with_ts (RtpSession * session, mblk_t *mp, uint32_t timestamp)
 	return error;
 }
 
+/**
+ *	Send the rtp datagram @mp to the destination set by rtp_session_set_remote_addr() 
+ *	with timestamp @timestamp. For audio data, the timestamp is the number
+ *	of the first sample resulting of the data transmitted. See rfc1889 for details.
+ *  The packet (@mp) is freed once it is sended.
+ *
+ *@param session a rtp session.
+ *@param mp a rtp packet presented as a mblk_t.
+ *@param timestamp the timestamp of the data to be sent.
+ * @return the number of bytes sent over the network.
+**/
+
+int rtp_session_sendm_with_ts(RtpSession *session, mblk_t *packet, uint32_t timestamp){
+	return __rtp_session_sendm_with_ts(session,packet,timestamp,timestamp);
+}
+
+
+
 
 /**
  *	Send a rtp datagram to the destination set by rtp_session_set_remote_addr() containing
@@ -855,10 +990,36 @@ extern void rtcp_parse(RtpSession *session, mblk_t *mp);
 
 
 static void payload_type_changed_notify(RtpSession *session, int paytype){
-	session->rcv.pt = paytype;
-	rtp_signal_table_emit (&session->on_payload_type_changed);	
+	PayloadType *pt = rtp_profile_get_payload(session->rcv.profile,paytype);
+	if (pt) {
+		session->rcv.pt = paytype;
+		rtp_signal_table_emit (&session->on_payload_type_changed);
+	}
 }
+/**
+ *	Try to get an rtp packet presented as a mblk_t structure from the rtp session at a given sequence number.
+ *	This function is very usefull for codec with Forward error correction capabilities
+ *
+ *	This function returns the entire packet (with header).
+ *
+ *	 *
+ * @param session a rtp session.
+ * @param sequence_number a sequence number.
+ *
+ * @return a rtp packet presented as a mblk_t, or NULL if not found.
+ **/
 
+mblk_t *
+rtp_session_pick_with_cseq (RtpSession * session, const uint16_t sequence_number) {
+	queue_t* q= &session->rtp.rq;
+	mblk_t* mb;
+	for (mb=qbegin(q); !qend(q,mb); mb=qnext(q,mb)){
+		if (rtp_get_seqnumber(mb)==sequence_number) {
+			return mb;
+		}
+	}
+	return NULL;
+}
 
 /**
  *	Try to get a rtp packet presented as a mblk_t structure from the rtp session.
@@ -868,6 +1029,23 @@ static void payload_type_changed_notify(RtpSession *session, int paytype){
  *	as it want. The RtpSession takes care of synchronisation between the stream timestamp
  *	and the user timestamp given here.
  *
+ *	This function returns the entire packet (with header).
+ *
+ *	The behaviour of this function has changed since version 0.15.0. Previously the payload data could be 
+ *	accessed using  mblk_t::b_cont::b_rptr field of the returned mblk_t.
+ *	This is no more the case.
+ *	The convenient way of accessing the payload data is to use rtp_get_payload() :
+ *	@code
+ *	unsigned char *payload;
+ *	int payload_size;
+ *	payload_size=rtp_get_payload(mp,&payload);
+ *	@endcode
+ *	OR simply skip the header this way, the data is then comprised between mp->b_rptr and mp->b_wptr:
+ *	@code
+ *	rtp_get_payload(mp,&mp->b_rptr);
+ *	@endcode
+ *
+ *
  * @param session a rtp session.
  * @param user_ts a timestamp.
  *
@@ -884,6 +1062,7 @@ rtp_session_recvm_with_ts (RtpSession * session, uint32_t user_ts)
 	RtpScheduler *sched=session->sched;
 	RtpStream *stream=&session->rtp;
 	int rejected=0;
+	bool_t read_socket=TRUE;
 
 	/* if we are scheduled, remember the scheduler time at which the application has
 	 * asked for its first timestamp */
@@ -894,7 +1073,7 @@ rtp_session_recvm_with_ts (RtpSession * session, uint32_t user_ts)
 		/* Set initial last_rcv_time to first recv time. */
 		if ((session->flags & RTP_SESSION_SEND_NOT_STARTED)
 		|| session->mode == RTP_SESSION_RECVONLY){
-			gettimeofday(&session->last_recv_time, NULL);
+			ortp_gettimeofday(&session->last_recv_time, NULL);
 		}
 		if (session->flags & RTP_SESSION_SCHEDULED)
 		{
@@ -902,10 +1081,17 @@ rtp_session_recvm_with_ts (RtpSession * session, uint32_t user_ts)
 			//ortp_message("setting snd_time_offset=%i",session->rtp.snd_time_offset);
 		}
 		rtp_session_unset_flag (session,RTP_SESSION_RECV_NOT_STARTED);
+	}else{
+		/*prevent reading from the sockets when two 
+		consecutives calls for a same timestamp*/
+		if (user_ts==session->rtp.rcv_last_app_ts)
+			read_socket=FALSE;
 	}
 	session->rtp.rcv_last_app_ts = user_ts;
-	rtp_session_rtp_recv (session, user_ts);
-	rtp_session_rtcp_recv(session);
+	if (read_socket){
+		rtp_session_rtp_recv (session, user_ts);
+		rtp_session_rtcp_recv(session);
+	}
 	/* check for telephone event first */
 	mp=getq(&session->rtp.tev_rq);
 	if (mp!=NULL){
@@ -932,22 +1118,21 @@ rtp_session_recvm_with_ts (RtpSession * session, uint32_t user_ts)
 		}
 		rtp = (rtp_header_t *) qfirst(q)->b_rptr;
 		session->rtp.rcv_ts_offset = rtp->timestamp;
-		/* remember the timestamp offset between the stream timestamp (random)
-		 * and the user timestamp, that very often starts at zero */
-		session->rtp.hwrcv_diff_ts = rtp->timestamp - user_ts;
-		/* remember the difference between the last received on the socket timestamp and the user timestamp */
-		session->rtp.rcv_diff_ts=session->rtp.hwrcv_diff_ts - session->rtp.jittctl.jitt_comp_ts;
 		session->rtp.rcv_last_ret_ts = user_ts;	/* just to have an init value */
 		session->rcv.ssrc = rtp->ssrc;
 		/* delete the recv synchronisation flag */
 		rtp_session_unset_flag (session, RTP_SESSION_RECV_SYNC);
 	}
+
 	/*calculate the stream timestamp from the user timestamp */
-	ts = user_ts + session->rtp.rcv_diff_ts;
-	if (session->permissive || session->rtp.jittctl.jitt_comp_ts==0)
-		mp = rtp_getq_permissive(&session->rtp.rq, ts,&rejected);
-	else
-		mp = rtp_getq(&session->rtp.rq, ts,&rejected);
+	ts = jitter_control_get_compensated_timestamp(&session->rtp.jittctl,user_ts);
+	if (session->rtp.jittctl.enabled==TRUE){
+		if (session->permissive)
+			mp = rtp_getq_permissive(&session->rtp.rq, ts,&rejected);
+		else{
+			mp = rtp_getq(&session->rtp.rq, ts,&rejected);
+		}
+	}else mp=getq(&session->rtp.rq);/*no jitter buffer at all*/
 	
 	stream->stats.outoftime+=rejected;
 	ortp_global_stats.outoftime+=rejected;
@@ -969,7 +1154,7 @@ rtp_session_recvm_with_ts (RtpSession * session, uint32_t user_ts)
 		{
 			payload_type_changed_notify(session, rtp->paytype);
 		}
-		/* patch the packet so that it has a timestamp compensated by the 
+		/* update the packet's timestamp so that it corrected by the 
 		adaptive jitter buffer mechanism */
 		if (session->rtp.jittctl.adaptive){
 			uint32_t changed_ts;
@@ -977,9 +1162,9 @@ rtp_session_recvm_with_ts (RtpSession * session, uint32_t user_ts)
 			timestamps*/
 			if (packet_ts!=session->rtp.rcv_last_ts)
 				jitter_control_update_corrective_slide(&session->rtp.jittctl);
-			changed_ts=packet_ts-session->rtp.jittctl.corrective_slide;
+			changed_ts=packet_ts+session->rtp.jittctl.corrective_slide;
 			rtp->timestamp=changed_ts;
-			/*printf("Returned packet has timestamp %u, with clock slide compensated it is %u\n",packet_ts,rtp->timestamp);*/
+			/*ortp_debug("Returned packet has timestamp %u, with clock slide compensated it is %u",packet_ts,rtp->timestamp);*/
 		}
 		session->rtp.rcv_last_ts = packet_ts;
 		if (!(session->flags & RTP_SESSION_FIRST_PACKET_DELIVERED)){
@@ -989,8 +1174,6 @@ rtp_session_recvm_with_ts (RtpSession * session, uint32_t user_ts)
 	else
 	{
 		ortp_debug ("No mp for timestamp queried");
-		stream->stats.unavaillable++;
-		ortp_global_stats.unavaillable++;
 	}
 	rtp_session_rtcp_process_recv(session);
 	
@@ -1000,13 +1183,14 @@ rtp_session_recvm_with_ts (RtpSession * session, uint32_t user_ts)
 		 * wanted expires */
 		/* but we must not block the process if the timestamp wanted by the application is older
 		 * than current time */
+		wait_point_lock(&session->rcv.wp);
 		packet_time =
 			rtp_session_ts_to_time (session,
 				     user_ts -
 				     session->rtp.rcv_query_ts_offset) +
 			session->rtp.rcv_time_offset;
 		ortp_debug ("rtp_session_recvm_with_ts: packet_time=%i, time=%i",packet_time, sched->time_);
-		wait_point_lock(&session->rcv.wp);
+		
 		if (TIME_IS_STRICTLY_NEWER_THAN (packet_time, sched->time_))
 		{
 			wait_point_wakeup_at(&session->rcv.wp,packet_time, (session->flags & RTP_SESSION_BLOCKING_MODE)!=0);
@@ -1019,38 +1203,6 @@ rtp_session_recvm_with_ts (RtpSession * session, uint32_t user_ts)
 }
 
 
-static int msg_to_buf (mblk_t * mp, uint8_t *buffer, int len)
-{
-	int rlen = len;
-	mblk_t *m, *mprev;
-	int mlen;
-	m = mp->b_cont;
-	mprev = mp;
-	while (m != NULL)
-	{
-		mlen = (int) (m->b_wptr - m->b_rptr);
-		if (mlen <= rlen)
-		{
-			mblk_t *consumed = m;
-			memcpy (buffer, m->b_rptr, mlen);
-			/* go to next mblk_t */
-			mprev->b_cont = m->b_cont;
-			m = m->b_cont;
-			consumed->b_cont = NULL;
-			freeb (consumed);
-			buffer += mlen;
-			rlen -= mlen;
-		}
-		else
-		{		/*if mlen>rlen */
-			memcpy (buffer, m->b_rptr, rlen);
-			m->b_rptr += rlen;
-			return len;
-		}
-	}
-	return len - rlen;
-}
-
 /**
  *	NOTE: use of this function is discouraged when sending payloads other than
  *	pcm/pcmu/pcma/adpcm types.
@@ -1091,125 +1243,40 @@ static int msg_to_buf (mblk_t * mp, uint8_t *buffer, int len)
  *
 **/
 int rtp_session_recv_with_ts (RtpSession * session, uint8_t * buffer,
-			       int len, uint32_t ts, int * have_more)
-{
-	mblk_t *mp;
-	int rlen = len;
-	int wlen, mlen;
-	uint32_t ts_int = 0;	/*the length of the data returned in the user supplied buffer, in TIMESTAMP UNIT */
-	PayloadType *payload;
-	RtpStream *stream=&session->rtp;
-
-	*have_more = 0;
-
-	mp = rtp_session_recvm_with_ts (session, ts);
-	payload =rtp_profile_get_payload (session->rcv.profile,
-					 session->rcv.pt);
-	if (payload==NULL){
-		ortp_warning("rtp_session_recv_with_ts: unable to recv an unsupported payload (%i)",session->rcv.pt);
-		if (mp!=NULL) freemsg(mp);
-		return -1;
+			       int len, uint32_t ts, int * have_more){
+	mblk_t *mp=NULL;
+	int plen,blen=0;
+	*have_more=0;
+	while(1){
+		if (session->pending){
+			mp=session->pending;
+			session->pending=NULL;
+		}else {
+			mp=rtp_session_recvm_with_ts(session,ts);
+			if (mp!=NULL) rtp_get_payload(mp,&mp->b_rptr);
+		}
+		if (mp){
+			plen=mp->b_wptr-mp->b_rptr;
+			if (plen<=len){
+				memcpy(buffer,mp->b_rptr,plen);
+				buffer+=plen;
+				blen+=plen;
+				len-=plen;
+				freemsg(mp);
+				mp=NULL;
+			}else{
+				memcpy(buffer,mp->b_rptr,len);
+				mp->b_rptr+=len;
+				buffer+=len;
+				blen+=len;
+				len=0;
+				session->pending=mp;
+				*have_more=1;
+				break;
+			}
+		}else break;
 	}
-	if (!(session->flags & RTP_SESSION_RECV_SYNC))
-	{
-		//ortp_debug("ts=%i   rcv_last_ret_ts=%i",ts,session->rtp.rcv_last_ret_ts);
-		if (RTP_TIMESTAMP_IS_STRICTLY_NEWER_THAN
-		    (ts, session->rtp.rcv_last_ret_ts))
-		{
-			/* the user has missed some data previously, so we are going to give him now. */
-			/* we must tell him to call the function once again with the same timestamp
-			 * by setting *have_more=1 */
-			*have_more = 1;
-		}
-		if (payload->type == PAYLOAD_AUDIO_CONTINUOUS)
-		{
-			ts_int = (len * payload->bits_per_sample) >> 3;
-			session->rtp.rcv_last_ret_ts += ts_int;
-			//ortp_debug("ts_int=%i",ts_int);
-		}
-		else
-			ts_int = 0;
-	}
-	else return 0;
-
-	/* try to fill the user buffer */
-	while (1)
-	{
-
-		if (mp != NULL)
-		{
-			mlen = msgdsize (mp->b_cont);
-			wlen = msg_to_buf (mp, buffer, rlen);
-			buffer += wlen;
-			rlen -= wlen;
-			ortp_debug("mlen=%i wlen=%i rlen=%i", mlen, wlen,
-				   rlen);
-			/* do we fill all the buffer ? */
-			if (rlen > 0)
-			{
-				/* we did not fill all the buffer */
-				freemsg (mp);
-				/* if we have continuous audio, try to get other packets to fill the buffer,
-				 * ie continue the loop */
-				//ortp_debug("User buffer not filled entirely");
-				if (ts_int > 0)
-				{
-					ts = session->rtp.rcv_last_ret_ts;
-					ortp_debug("Need more: will ask for %i.", ts);
-				}
-				else
-					return len - rlen;
-			}
-			else if (mlen > wlen)
-			{
-				int unread =
-					mlen - wlen + (int) (mp->b_wptr -
-						       mp->b_rptr);
-				/* not enough space in the user supplied buffer */
-				/* we re-enqueue the msg with its updated read pointers for next time */
-				ortp_debug ("Re-enqueuing packet.");
-				rtp_putq (&session->rtp.rq, mp);
-				/* quite ugly: I change the stats ... */
-				ortp_global_stats.recv -= unread;
-				stream->stats.recv -= unread;
-				return len;
-			}
-			else
-			{
-				/* the entire packet was written to the user buffer */
-				freemsg (mp);
-				return len;
-			}
-		}
-		else
-		{
-			/* fill with a zero pattern (silence) */
-			if (payload->pattern_length != 0)
-			{
-				int i = 0, j = 0;
-				while (i < rlen)
-				{
-					buffer[i] = payload->zero_pattern[j];
-					i++;
-					j++;
-					if (j <= payload->pattern_length)
-						j = 0;
-				}
-				return len;
-			}
-			*have_more = 0;
-			return 0;
-		}
-		mp = rtp_session_recvm_with_ts (session, ts);
-		payload = rtp_profile_get_payload (session->rcv.profile,
-						 session->rcv.pt);
-		if (payload==NULL){
-			ortp_warning("rtp_session_recv_with_ts: unable to recv an unsupported payload.");
-			if (mp!=NULL) freemsg(mp);
-			return -1;
-		}
-	}
-	return -1;
+	return blen;
 }
 /**
  *	When the rtp session is scheduled and has started to send packets, this function
@@ -1288,11 +1355,17 @@ void rtp_session_set_time_jump_limit(RtpSession *session, int milisecs){
  * Closes the rtp and rtcp sockets.
 **/
 void rtp_session_release_sockets(RtpSession *session){
-	if (session->rtp.socket>=0) close_socket (session->rtp.socket);
-	if (session->rtcp.socket>=0) close_socket (session->rtcp.socket);
+	if (session->rtp.socket!=(ortp_socket_t)-1) close_socket (session->rtp.socket);
+	if (session->rtcp.socket!=(ortp_socket_t)-1) close_socket (session->rtcp.socket);
 	session->rtp.socket=-1;
 	session->rtcp.socket=-1;
+
+	if (session->rtp.tr && session->rtp.tr->t_close)
+		session->rtp.tr->t_close(session->rtp.tr, session->rtp.tr->data);
 	session->rtp.tr = 0;
+
+	if (session->rtcp.tr && session->rtcp.tr->t_close)
+		session->rtcp.tr->t_close(session->rtcp.tr, session->rtcp.tr->data);
 	session->rtcp.tr = 0;
 
 	/* don't discard remote addresses, then can be preserved for next use.
@@ -1354,6 +1427,57 @@ void rtp_session_uninit (RtpSession * session)
 	if (session->sd!=NULL) freemsg(session->sd);
 
 	session->signal_tables = o_list_free(session->signal_tables);
+	msgb_allocator_uninit(&session->allocator);
+
+	if (session->net_sim_ctx)
+		ortp_network_simulator_destroy(session->net_sim_ctx);
+
+#if (_WIN32_WINNT >= 0x0600) && !WINAPI_FAMILY_APP
+	if (session->rtp.QoSFlowID != 0)
+    {
+		OSVERSIONINFOEX ovi;
+		memset(&ovi, 0, sizeof(ovi));
+		ovi.dwOSVersionInfoSize = sizeof(ovi);
+		GetVersionEx((LPOSVERSIONINFO) & ovi);
+
+		ortp_message("check OS support for qwave.lib: %i %i %i\n",
+					ovi.dwMajorVersion, ovi.dwMinorVersion, ovi.dwBuildNumber);
+		if (ovi.dwMajorVersion > 5) {
+
+			if (FAILED(__HrLoadAllImportsForDll("qwave.dll"))) {
+				ortp_warning("Failed to load qwave.dll: no QoS available\n" );
+			}
+			else
+			{
+				BOOL QoSResult;
+				QoSResult = QOSRemoveSocketFromFlow(session->rtp.QoSHandle, 
+													0, 
+													session->rtp.QoSFlowID, 
+													0);
+
+				if (QoSResult != TRUE){
+					ortp_error("QOSRemoveSocketFromFlow failed to end a flow with error %d\n", 
+							GetLastError());
+				}
+				session->rtp.QoSFlowID=0;
+			}
+		}
+    }
+
+    if (session->rtp.QoSHandle != NULL)
+    {
+        QOSCloseHandle(session->rtp.QoSHandle);
+		session->rtp.QoSHandle=NULL;
+    }
+#endif
+}
+
+/**
+ * Sets the number of packets containg a new SSRC that will trigger the
+ * "ssrc_changed" callback.
+**/
+void rtp_session_set_ssrc_changed_threshold(RtpSession *session, int numpackets){
+	session->rtp.ssrc_changed_thres=numpackets;
 }
 
 /**
@@ -1387,14 +1511,17 @@ void rtp_session_reset (RtpSession * session)
 	session->rtp.rcv_time_offset = 0;
 	session->rtp.rcv_ts_offset = 0;
 	session->rtp.rcv_query_ts_offset = 0;
-	session->rtp.rcv_diff_ts = 0;
 	session->rtp.rcv_last_ts = 0;
 	session->rtp.rcv_last_app_ts = 0;
 	session->rtp.hwrcv_extseq = 0;
 	session->rtp.hwrcv_since_last_SR=0;
 	session->rtp.snd_seq = 0;
+	session->rtp.sent_payload_bytes=0;
+	rtp_session_clear_send_error_code(session);
+	rtp_session_clear_recv_error_code(session);
 	rtp_stats_reset(&session->rtp.stats);
 	rtp_session_resync(session);
+	session->ssrc_set=FALSE;
 }
 
 /**
@@ -1404,6 +1531,63 @@ const rtp_stats_t * rtp_session_get_stats(const RtpSession *session){
 	return &session->rtp.stats;
 }
 
+/**
+ * Retrieves the session's jitter specific statistics.
+**/
+const jitter_stats_t * rtp_session_get_jitter_stats( const RtpSession *session ) {
+	return &session->rtp.jitter_stats;
+}
+
+/**
+ * @brief For <b>test purpose only</b>, sets a constant lost packet value within <b>all</b> RTCP output packets.@n
+ *
+ * The SR or RR RTCP packet contain a lost packet field. After this procedure is called, the lost packet field will be set to a constant value in all output SR or RR packets. This parameter will overridden the actual number of lost packets in the input RTP stream that the RTCP stack had previously processed.
+ * @param s : the rtp session.
+ * @param value : the lost packets test vector value.
+**/
+void rtp_session_rtcp_set_lost_packet_value( struct _RtpSession *s, const unsigned int value ) {
+	s->lost_packets_test_vector = value;
+	s->flags|=RTCP_OVERRIDE_LOST_PACKETS;
+}
+
+/**
+ * @brief For <b>test purpose only</b>, sets a constant interarrival_jitter value within <b>all</b> RTCP output packets.@n
+ *
+ * The SR or RR RTCP packet contain an interarrival jitter field. After this procedure is called, the interarrival jitter field will be set to a constant value in all output SR or RR packets. This parameter will overridden the actual interarrival jitter value that was processed by the RTCP stack.
+ * @param s : the rtp session.
+ * @param value : the interarrival jitter test vector value.
+**/
+void rtp_session_rtcp_set_jitter_value( struct _RtpSession *s, const unsigned int value ) {
+	s->interarrival_jitter_test_vector = value;
+	s->flags|=RTCP_OVERRIDE_JITTER;
+}
+
+/**
+ * @brief For <b>test purpose only</b>, simulates a constant RTT (Round Trip Time) value by setting the LSR field within <b>all</b> returned RTCP output packets.@n
+ *
+ * The RTT processing involves two RTCP packets exchanged between two different devices.@n
+ * In a <b>normal</b> operation the device 1 issues a SR packets at time T0, hence this packet has a timestamp field set to T0.
+ * The LSR and DLSR fiels of that packet are not considered here. This packet is received by the Device 2 at T1. 
+ * In response, the Device 2 issues another SR or RR packets at T2 with the following fields;
+ * - a timestamp set to T2.
+ * - a LSR (Last SR packet timestamp) field set to T0 ( this value has been extracted from the first packet).
+ * - a DLSR (Delay since Last SR packet) field set to (T2 - T1).
+ * .
+ * This packet is received by The Device 1 at T3. So the Device 1 is now able to process the RTT using the formula :
+ * RTT = T3 - LSR - DLSR = (T1 - T0) - (T3 - T2).@n
+ * This way of processing is described in par. 6.4 of the RFC3550 standard.
+ *
+ * In the <b>test</b> mode that is enabled by this procedure, the RTCP stack is considered as beeing part of the device 2. For setting the RTT to a constant RTT0 value, the Device 2 artificially sets the LSR field of the second packet to (T1 - RTT0), instead of T0 in normal mode. The two other fields (timestamp and DLSR) are set as in the normal mode. So the Device 1 will process :
+ * RTT = T3 - LSR - DLSR = RTT0 + (T3 - T2) that is near to RTT0 is T3 - T2 is small enough.
+ * @note It is impossible to actually make the mesured RTT strictly equal to RTT0, as the packet trip time (T3 - T2) is unknown when this packet is issued by the Device 2.
+ * @param s : the rtp session.
+ * @param value : The desired RTT test vector value (RTT0).
+**/
+void rtp_session_rtcp_set_delay_value( struct _RtpSession *s, const unsigned int value ) {
+	s->delay_test_vector= value;
+	s->flags|=RTCP_OVERRIDE_DELAY;
+}
+
 void rtp_session_reset_stats(RtpSession *session){
 	memset(&session->rtp.stats,0,sizeof(rtp_stats_t));
 }
@@ -1463,27 +1647,62 @@ static float compute_bw(struct timeval *orig, unsigned int bytes){
 	float bw;
 	float time;
 	if (bytes==0) return 0;
-	gettimeofday(&current,NULL);
-	time=(current.tv_sec - orig->tv_sec) +
-		((current.tv_usec - orig->tv_usec)*1e-6);
-	bw=((float)bytes)*8/time;
+	ortp_gettimeofday(&current,NULL);
+	time=(float)(current.tv_sec - orig->tv_sec) +
+		((float)(current.tv_usec - orig->tv_usec)*1e-6);
+	bw=((float)bytes)*8/(time+0.001); 
+	/*+0.0001 avoids a division by zero without changing the results significatively*/
 	return bw;
 }
 
 float rtp_session_compute_recv_bandwidth(RtpSession *session){
-	float bw;
-	bw=compute_bw(&session->rtp.recv_bw_start,session->rtp.recv_bytes);
-	session->rtp.recv_bytes=0;
+	float bw=-1;
+	if (session) {
+		bw=compute_bw(&session->rtp.recv_bw_start,session->rtp.recv_bytes);
+		session->rtp.recv_bytes=0;
+	}
 	return bw;
 }
 
 float rtp_session_compute_send_bandwidth(RtpSession *session){
-	float bw;
-	bw=compute_bw(&session->rtp.send_bw_start,session->rtp.sent_bytes);
-	session->rtp.sent_bytes=0;
+	float bw=-1;
+	if (session) {
+		bw=compute_bw(&session->rtp.send_bw_start,session->rtp.sent_bytes);
+		session->rtp.sent_bytes=0;
+	}
 	return bw;
 }
 
+int rtp_session_get_last_send_error_code(RtpSession *session){
+	return session->rtp.send_errno;
+}
+
+void rtp_session_clear_send_error_code(RtpSession *session){
+	session->rtp.send_errno=0;
+}
+
+int rtp_session_get_last_recv_error_code(RtpSession *session){
+	return session->rtp.recv_errno;
+}
+
+void rtp_session_clear_recv_error_code(RtpSession *session){
+	session->rtp.send_errno=0;
+}
+
+/**
+ * Returns the last known round trip propagation delay.
+ *
+ * This value is known after successful RTCP SR or RR exchanged between a sender and a receiver.
+ * oRTP automatically takes care of sending SR or RR packets.
+ * You might want to call this function when you receive an RTCP event (see rtp_session_register_event_queue() ).
+ * This value might not be known: at the beginning when no RTCP packets have been exchanged yet, or simply because the
+ * rtcp channel is broken due to firewall problematics, or because the remote implementation does not support RTCP.
+ *
+ * @returns the round trip propagation time in seconds if known, -1 if unknown. 
+**/
+float rtp_session_get_round_trip_propagation(RtpSession *session){
+	return session->rtt;
+}
 
 /**
  * Destroys a rtp session.
@@ -1512,6 +1731,68 @@ void rtp_add_csrc(mblk_t *mp, uint32_t csrc)
 	hdr->cc++;
 }
 
+/**
+ * Get a pointer to the beginning of the payload data of the RTP packet.
+ * @param packet a RTP packet represented as a mblk_t
+ * @param start a pointer to the beginning of the payload data, pointing inside the packet.
+ * @return the length of the payload data.
+**/
+int rtp_get_payload(mblk_t *packet, unsigned char **start){
+	unsigned char *tmp;
+	int header_len=RTP_FIXED_HEADER_SIZE+(rtp_get_cc(packet)*4);
+	tmp=packet->b_rptr+header_len;
+	if (tmp>packet->b_wptr){
+		if (packet->b_cont!=NULL){
+			tmp=packet->b_cont->b_rptr+(header_len- (packet->b_wptr-packet->b_rptr));
+			if (tmp<=packet->b_cont->b_wptr){
+				*start=tmp;
+				return packet->b_cont->b_wptr-tmp;
+			}
+		}
+		ortp_warning("Invalid RTP packet");
+		return -1;
+	}
+	if (rtp_get_extbit(packet)){
+		int extsize=rtp_get_extheader(packet,NULL,NULL);
+		if (extsize>=0){
+			tmp+=4+extsize;
+		}
+	}
+	*start=tmp;
+	return packet->b_wptr-tmp;
+}
+
+/**
+ * Obtain the extension header if any.
+ * @param packet the RTP packet.
+ * @param profile the profile field of the extension header
+ * @param start_ext_header pointer that will be set to the beginning of the payload of the extension header.
+ * @return the size of the extension in bytes (the payload size, it can be 0), -1 if parsing of the extension header failed or if no extension is present.
+**/
+int rtp_get_extheader(mblk_t *packet, uint16_t *profile, uint8_t **start_ext){
+	int size=0;
+	uint8_t *ext_header;
+	if (rtp_get_extbit(packet)){
+		ext_header=packet->b_rptr+RTP_FIXED_HEADER_SIZE+(rtp_get_cc(packet)*4);
+		if (ext_header+4 <= packet->b_wptr){
+			uint32_t h=ntohl(*(uint32_t*)ext_header);
+			size=(int)(h & 0xFFFF);
+			if (profile) *profile=(h>>16);
+			size=(size*4); /*the size is given in the packet as multiple of 32 bit words, excluding the 4 byte header*/
+			if ((ext_header+4+size)> packet->b_wptr){
+				ortp_warning("Inconsistent size for rtp extension header");
+				return -1;
+			}
+			if (start_ext) *start_ext=ext_header+4;
+			return size;
+		}else{
+			ortp_warning("Insufficient size for rtp extension header.");
+			return -1;
+		}
+	}
+	return -1;
+}
+
 
 /**
  *  Gets last time a valid RTP or RTCP packet was received.
@@ -1522,7 +1803,11 @@ void rtp_add_csrc(mblk_t *mp, uint32_t csrc)
 void
 rtp_session_get_last_recv_time(RtpSession *session, struct timeval *tv)
 {
-    *tv = session->last_recv_time;
+#ifdef PERF
+	ortp_error("rtp_session_get_last_recv_time() feature disabled.");
+#else
+    	*tv = session->last_recv_time;
+#endif
 }
 
 
@@ -1580,3 +1865,6 @@ void rtp_session_process (RtpSession * session, uint32_t time, RtpScheduler *sch
 	wait_point_unlock(&session->rcv.wp);
 }
 
+void rtp_session_set_reuseaddr(RtpSession *session, bool_t yes) {
+	session->reuseaddr=yes;
+}
