@@ -17,22 +17,30 @@
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
+
+#define _GNU_SOURCE
+
+#if defined(WIN32) || defined(_WIN32_WCE)
+#include "ortp-config-win32.h"
+#elif HAVE_CONFIG_H
+#include "ortp-config.h" /*needed for HAVE_SYS_UIO_H and HAVE_ARC4RANDOM */
+#endif
 #include "ortp/ortp.h"
 #include "utils.h"
 #include "ortp/rtpsession.h"
 #include "rtpsession_priv.h"
 
-#if defined(WIN32) || defined(_WIN32_WCE)
-#include "ortp-config-win32.h"
-#elif HAVE_CONFIG_H
-#include "ortp-config.h" /*needed for HAVE_SYS_UIO_H */
-#endif
-
 #if (_WIN32_WINNT >= 0x0600)
 #include <delayimp.h>
 #undef ExternC
+#ifndef WINAPI_FAMILY_PHONE_APP
 #include <QOS2.h>
 #endif
+#endif
+
+#if (defined(WIN32) || defined(_WIN32_WCE)) && !defined(WINAPI_FAMILY_PHONE_APP)
+#include <mswsock.h>
+#endif
 
 #ifdef HAVE_SYS_UIO_H
 #include <sys/uio.h>
@@ -41,6 +49,37 @@
 
 #define can_connect(s)	( (s)->use_connect && !(s)->symmetric_rtp)
 
+#if defined(WIN32) || defined(_WIN32_WCE)
+#ifndef WSAID_WSARECVMSG
+/* http://source.winehq.org/git/wine.git/blob/HEAD:/include/mswsock.h */
+#define WSAID_WSARECVMSG {0xf689d7c8,0x6f1f,0x436b,{0x8a,0x53,0xe5,0x4f,0xe3,0x51,0xc3,0x22}}
+#define MAX_NATURAL_ALIGNMENT sizeof(DWORD)
+#define TYPE_ALIGNMENT(t) FIELD_OFFSET(struct { char x; t test; },test)
+typedef WSACMSGHDR *LPWSACMSGHDR;
+#define WSA_CMSGHDR_ALIGN(length) (((length) + TYPE_ALIGNMENT(WSACMSGHDR)-1) & (~(TYPE_ALIGNMENT(WSACMSGHDR)-1)))
+#define WSA_CMSGDATA_ALIGN(length) (((length) + MAX_NATURAL_ALIGNMENT-1) & (~(MAX_NATURAL_ALIGNMENT-1)))
+#define WSA_CMSG_FIRSTHDR(msg) (((msg)->Control.len >= sizeof(WSACMSGHDR)) ? (LPWSACMSGHDR)(msg)->Control.buf : (LPWSACMSGHDR)NULL)
+#define WSA_CMSG_NXTHDR(msg,cmsg) ((!(cmsg)) ? WSA_CMSG_FIRSTHDR(msg) : ((((u_char *)(cmsg) + WSA_CMSGHDR_ALIGN((cmsg)->cmsg_len) + sizeof(WSACMSGHDR)) > (u_char *)((msg)->Control.buf) + (msg)->Control.len) ? (LPWSACMSGHDR)NULL : (LPWSACMSGHDR)((u_char *)(cmsg) + WSA_CMSGHDR_ALIGN((cmsg)->cmsg_len))))
+#define WSA_CMSG_DATA(cmsg) ((u_char *)(cmsg) + WSA_CMSGDATA_ALIGN(sizeof(WSACMSGHDR)))
+#endif
+#undef CMSG_FIRSTHDR
+#define CMSG_FIRSTHDR WSA_CMSG_FIRSTHDR
+#undef CMSG_NXTHDR
+#define CMSG_NXTHDR WSA_CMSG_NXTHDR
+#undef CMSG_DATA
+#define CMSG_DATA WSA_CMSG_DATA
+typedef INT  (WINAPI * LPFN_WSARECVMSG)(SOCKET, LPWSAMSG, LPDWORD, LPWSAOVERLAPPED, LPWSAOVERLAPPED_COMPLETION_ROUTINE);
+static LPFN_WSARECVMSG ortp_WSARecvMsg = NULL;
+
+#endif
+
+#if defined(WIN32) || defined(_WIN32_WCE) || defined(__QNX__)
+/* Mingw32 does not define AI_V4MAPPED, however it is supported starting from Windows Vista. QNX also does not define AI_V4MAPPED. */
+#	ifndef AI_V4MAPPED
+#	define AI_V4MAPPED 0x00000800
+#	endif
+#endif
+
 static bool_t try_connect(int fd, const struct sockaddr *dest, socklen_t addrlen){
 	if (connect(fd,dest,addrlen)<0){
 		ortp_warning("Could not connect() socket: %s",getSocketError());
@@ -86,11 +125,19 @@ static ortp_socket_t create_and_bind(const char *addr, int port, int *sock_famil
 				ortp_warning ("Fail to set rtp address reusable: %s.", getSocketError());
 			}
 		}
+#if defined(ORTP_TIMESTAMP)
+		err = setsockopt (sock, SOL_SOCKET, SO_TIMESTAMP,
+			(SOCKET_OPTION_VALUE)&optval, sizeof (optval));
+		if (err < 0)
+		{
+			ortp_warning ("Fail to set rtp timestamp: %s.",getSocketError());
+		}
+#endif
 
 		*sock_family=res->ai_family;
 		err = bind (sock, res->ai_addr, res->ai_addrlen);
 		if (err != 0){
-			ortp_warning ("Fail to bind rtp socket to (addr=%s port=%i) : %s.", addr,port, getSocketError());
+			ortp_debug ("Fail to bind rtp socket to (addr=%s port=%i) : %s.", addr,port, getSocketError());
 			close_socket (sock);
 			sock=-1;
 			continue;
@@ -156,6 +203,14 @@ static ortp_socket_t create_and_bind(const char *addr, int port, int *sock_famil
 			ortp_warning ("Fail to set rtp address reusable: %s.",getSocketError());
 		}
 	}
+#if defined(ORTP_TIMESTAMP)
+	err = setsockopt (sock, SOL_SOCKET, SO_TIMESTAMP,
+			(SOCKET_OPTION_VALUE)&optval, sizeof (optval));
+	if (err < 0)
+	{
+		ortp_warning ("Fail to set rtp timestamp: %s.",getSocketError());
+	}
+#endif
 
 	err = bind (sock,
 		    (struct sockaddr *) &saddr,
@@ -163,10 +218,20 @@ static ortp_socket_t create_and_bind(const char *addr, int port, int *sock_famil
 
 	if (err != 0)
 	{
-		ortp_warning ("Fail to bind rtp socket to port %i: %s.", port, getSocketError());
+		ortp_debug ("Fail to bind rtp socket to port %i: %s.", port, getSocketError());
 		close_socket (sock);
 		return -1;
 	}
+#endif
+#if defined(WIN32) || defined(_WIN32_WCE)
+	if (ortp_WSARecvMsg == NULL) {
+		GUID guid = WSAID_WSARECVMSG;
+		DWORD bytes_returned;
+		if (WSAIoctl(sock, SIO_GET_EXTENSION_FUNCTION_POINTER, &guid, sizeof(guid),
+			&ortp_WSARecvMsg, sizeof(ortp_WSARecvMsg), &bytes_returned, NULL, NULL) == SOCKET_ERROR) {
+			ortp_warning("WSARecvMsg function not found.");
+		}
+	}
 #endif
 	if (sock!=-1){
 		set_non_blocking_socket (sock);
@@ -215,11 +280,16 @@ static ortp_socket_t create_and_bind_random(const char *localip, int *sock_famil
 	for (retry=0;retry<100;retry++)
 	{
 		int localport;
+#ifdef HAVE_ARC4RANDOM
+		localport = 5000 + (int)arc4random_uniform(0x10000 - 5000);
+		localport &= 0xfffe;
+#else
 		do
 		{
 			localport = (rand () + 5000) & 0xfffe;
 		}
 		while ((localport < 5000) || (localport > 0xffff));
+#endif
 		/*do not set REUSEADDR in case of random allocation */
 		sock = create_and_bind(localip, localport, sock_family,FALSE);
 		if (sock!=-1) {
@@ -235,7 +305,8 @@ static ortp_socket_t create_and_bind_random(const char *localip, int *sock_famil
  *rtp_session_set_local_addr:
  *@session:		a rtp session freshly created.
  *@addr:		a local IP address in the xxx.xxx.xxx.xxx form.
- *@port:		a local port or -1 to let oRTP choose the port randomly
+ *@rtp_port:		a local port or -1 to let oRTP choose the port randomly
+ *@rtcp_port:		a local port or -1 to let oRTP choose the port randomly
  *
  *	Specify the local addr to be use to listen for rtp packets or to send rtp packet from.
  *	In case where the rtp session is send-only, then it is not required to call this function:
@@ -247,37 +318,40 @@ static ortp_socket_t create_and_bind_random(const char *localip, int *sock_famil
 **/
 
 int
-rtp_session_set_local_addr (RtpSession * session, const char * addr, int port)
+rtp_session_set_local_addr (RtpSession * session, const char * addr, int rtp_port, int rtcp_port)
 {
 	ortp_socket_t sock;
 	int sockfamily;
-	bool_t reuse_addr;
 	if (session->rtp.socket!=(ortp_socket_t)-1){
 		/* don't rebind, but close before*/
 		rtp_session_release_sockets(session);
 	}
-#ifdef SO_REUSE_ADDR
-	reuse_addr=TRUE;
-#else
-	reuse_addr=FALSE;
-#endif
 	/* try to bind the rtp port */
-	if (port>0)
-		sock=create_and_bind(addr,port,&sockfamily,reuse_addr);
+	if (rtp_port>0)
+		sock=create_and_bind(addr,rtp_port,&sockfamily,session->reuseaddr);
 	else
-		sock=create_and_bind_random(addr,&sockfamily,&port);
+		sock=create_and_bind_random(addr,&sockfamily,&rtp_port);
 	if (sock!=-1){
 		set_socket_sizes(sock,session->rtp.snd_socket_size,session->rtp.rcv_socket_size);
 		session->rtp.sockfamily=sockfamily;
 		session->rtp.socket=sock;
-		session->rtp.loc_port=port;
+		session->rtp.loc_port=rtp_port;
 		/*try to bind rtcp port */
-		sock=create_and_bind(addr,port+1,&sockfamily,reuse_addr);
+		if (rtcp_port<0) {
+			rtcp_port=rtp_port+1;
+			sock=create_and_bind(addr,rtcp_port,&sockfamily,session->reuseaddr);
+			if (sock==(ortp_socket_t)-1) {
+				sock=create_and_bind_random(addr,&sockfamily,&rtcp_port);
+			}
+		} else {
+			sock=create_and_bind(addr,rtcp_port,&sockfamily,session->reuseaddr);
+		}
 		if (sock!=(ortp_socket_t)-1){
 			session->rtcp.sockfamily=sockfamily;
 			session->rtcp.socket=sock;
-		}else{
-			ortp_warning("Could not create and bind rtcp socket.");
+		}else {
+			ortp_debug("Could not create and bind rtcp socket.");
+			return -1;
 		}
 		
 		/* set socket options (but don't change chosen states) */
@@ -287,11 +361,73 @@ rtp_session_set_local_addr (RtpSession * session, const char * addr, int port)
 
 		return 0;
 	}
-	ortp_error("Could not bind RTP socket on port to %s port %i",addr,port);
+	ortp_debug("Could not bind RTP socket on port to %s port %i",addr,rtp_port);
 	return -1;
 }
 
 
+/**
+ *rtp_session_set_pktinfo:
+ *@session: a rtp session
+ *@activate: activation flag (0 to deactivate, other value to activate)
+ *
+ * (De)activates packet info for incoming and outgoing packets.
+ *
+ * Returns: 0 on success.
+ *
+**/
+int rtp_session_set_pktinfo(RtpSession *session, int activate)
+{
+	int retval;
+	int optname;
+#if defined(WIN32) || defined(_WIN32_WCE)
+	char optval[sizeof(DWORD)];
+	int optlen = sizeof(optval);
+#else
+	int *optval = &activate;
+	int optlen = sizeof(activate);
+#endif
+
+	// Dont't do anything if socket hasn't been created yet
+	if (session->rtp.socket == (ortp_socket_t)-1) return 0;
+
+#if defined(WIN32) || defined(_WIN32_WCE)
+	memset(optval, activate, sizeof(optval));
+#endif
+
+	switch (session->rtp.sockfamily) {
+		case AF_INET:
+#ifdef IP_PKTINFO
+			optname = IP_PKTINFO;
+#else
+			optname = IP_RECVDSTADDR;
+#endif
+			retval = setsockopt(session->rtp.socket, IPPROTO_IP, optname, optval, optlen);
+			if (retval < 0) break;
+			retval = setsockopt(session->rtcp.socket, IPPROTO_IP, optname, optval, optlen);
+			break;
+#ifdef ORTP_INET6
+		case AF_INET6:
+#ifdef IPV6_PKTINFO
+			optname = IPV6_PKTINFO;
+#else
+			optname = IPV6_RECVDSTADDR;
+#endif
+			retval = setsockopt(session->rtp.socket, IPPROTO_IPV6, optname, optval, optlen);
+			if (retval < 0) break;
+			retval = setsockopt(session->rtcp.socket, IPPROTO_IPV6, optname, optval, optlen);
+			break;
+#endif
+		default:
+			retval = -1;
+			break;
+	}
+
+	if (retval < 0) ortp_warning("Failed to set packet info on socket.");
+	return retval;
+}
+
+
 /**
  *rtp_session_set_multicast_ttl:
  *@session: a rtp session
@@ -450,6 +586,8 @@ int rtp_session_get_multicast_loopback(RtpSession *session)
 int rtp_session_set_dscp(RtpSession *session, int dscp){
 	int retval=0;
 	int tos;
+	int proto;
+	int value_type;
 #if (_WIN32_WINNT >= 0x0600)
 	OSVERSIONINFOEX ovi;
 #endif
@@ -460,7 +598,7 @@ int rtp_session_set_dscp(RtpSession *session, int dscp){
 	// Don't do anything if socket hasn't been created yet
 	if (session->rtp.socket == (ortp_socket_t)-1) return 0;
 
-#if (_WIN32_WINNT >= 0x0600)
+#if (_WIN32_WINNT >= 0x0600) && !defined(WINAPI_FAMILY_PHONE_APP)
 	memset(&ovi, 0, sizeof(ovi));
 	ovi.dwOSVersionInfoSize = sizeof(ovi);
 	GetVersionEx((LPOSVERSIONINFO) & ovi);
@@ -523,30 +661,34 @@ int rtp_session_set_dscp(RtpSession *session, int dscp){
 		tos = (session->dscp << 2) & 0xFC;
 		switch (session->rtp.sockfamily) {
 			case AF_INET:
-			retval = setsockopt(session->rtp.socket, IPPROTO_IP, IP_TOS, (SOCKET_OPTION_VALUE)&tos, sizeof(tos));
+				proto=IPPROTO_IP;
+				value_type=IP_TOS;
 			break;
 #ifdef ORTP_INET6
 		case AF_INET6:
+			proto=IPPROTO_IPV6;
 #	ifdef IPV6_TCLASS /*seems not defined by my libc*/
-			retval = setsockopt(session->rtp.socket, IPPROTO_IPV6, IPV6_TCLASS,
-			 (SOCKET_OPTION_VALUE)&tos, sizeof(tos));
+			value_type=IPV6_TCLASS;
 #	else
-			/*in case that works:*/
-			retval = setsockopt(session->rtp.socket, IPPROTO_IPV6, IP_TOS,
-			 (SOCKET_OPTION_VALUE)&tos, sizeof(tos));
+			value_type=IP_TOS;
 #	endif
 			break;
 #endif
 		default:
-			retval=-1;
+			ortp_error("Cannot set DSCP because socket family is unspecified.");
+			return -1;
 		}
-#if (_WIN32_WINNT >= 0x0600)
+		retval = setsockopt(session->rtp.socket, proto, value_type, (SOCKET_OPTION_VALUE)&tos, sizeof(tos));
+		if (retval==-1)
+			ortp_error("Fail to set DSCP value on rtp socket: %s",getSocketError());
+		if (session->rtcp.socket != (ortp_socket_t)-1){
+			if (setsockopt(session->rtcp.socket, proto, value_type, (SOCKET_OPTION_VALUE)&tos, sizeof(tos))==-1){
+				ortp_error("Fail to set DSCP value on rtcp socket: %s",getSocketError());
+			}
+		}
+#if (_WIN32_WINNT >= 0x0600) && !defined(WINAPI_FAMILY_PHONE_APP)
 	}
 #endif
-
-	if (retval<0)
-		ortp_warning("Failed to set DSCP value on socket.");
-
 	return retval;
 }
 
@@ -609,14 +751,15 @@ static char * ortp_inet_ntoa(struct sockaddr *addr, int addrlen, char *dest, int
 **/
 int
 rtp_session_set_remote_addr (RtpSession * session, const char * addr, int port){
-	return rtp_session_set_remote_addr_full (session, addr, port, port+1);
+	return rtp_session_set_remote_addr_full (session, addr, port, addr, port+1);
 }
 
 /**
  *rtp_session_set_remote_addr_full:
  *@session:		a rtp session freshly created.
- *@addr:		a local IP address in the xxx.xxx.xxx.xxx form.
+ *@rtp_addr:		a local IP address in the xxx.xxx.xxx.xxx form.
  *@rtp_port:		a local rtp port.
+ *@rtcp_addr:		a local IP address in the xxx.xxx.xxx.xxx form.
  *@rtcp_port:		a local rtcp port.
  *
  *	Sets the remote address of the rtp session, ie the destination address where rtp packet
@@ -627,17 +770,20 @@ rtp_session_set_remote_addr (RtpSession * session, const char * addr, int port){
 **/
 
 int
-rtp_session_set_remote_addr_full (RtpSession * session, const char * addr, int rtp_port, int rtcp_port)
+rtp_session_set_remote_addr_full (RtpSession * session, const char * rtp_addr, int rtp_port, const char * rtcp_addr, int rtcp_port)
 {
 	int err;
 #ifdef ORTP_INET6
 	struct addrinfo hints, *res0, *res;
 	char num[8];
+	
 	memset(&hints, 0, sizeof(hints));
-	hints.ai_family = PF_UNSPEC;
+	hints.ai_family = (session->rtp.socket == -1) ? AF_UNSPEC : session->rtp.sockfamily;
 	hints.ai_socktype = SOCK_DGRAM;
+	hints.ai_flags = hints.ai_family==AF_INET6 ? AI_V4MAPPED : 0;
+	
 	snprintf(num, sizeof(num), "%d", rtp_port);
-	err = getaddrinfo(addr, num, &hints, &res0);
+	err = getaddrinfo(rtp_addr, num, &hints, &res0);
 	if (err) {
 		ortp_warning ("Error in socket address: %s", gai_strerror(err));
 		return -1;
@@ -649,10 +795,10 @@ rtp_session_set_remote_addr_full (RtpSession * session, const char * addr, int r
 #ifdef ORTP_INET6
 		/* bind to an address type that matches the destination address */
 		if (res0->ai_addr->sa_family==AF_INET6)
-			err = rtp_session_set_local_addr (session, "::", -1);
-		else err=rtp_session_set_local_addr (session, "0.0.0.0", -1);
+			err = rtp_session_set_local_addr (session, "::", -1, -1);
+		else err=rtp_session_set_local_addr (session, "0.0.0.0", -1, -1);
 #else
-		err = rtp_session_set_local_addr (session, "0.0.0.0", -1);
+		err = rtp_session_set_local_addr (session, "0.0.0.0", -1, -1);
 #endif
 		if (err<0) return -1;
 	}
@@ -670,15 +816,16 @@ rtp_session_set_remote_addr_full (RtpSession * session, const char * addr, int r
 	}
 	freeaddrinfo(res0);
 	if (err) {
-		ortp_warning("Could not set destination for RTP socket to %s:%i.",addr,rtp_port);
+		ortp_warning("Could not set destination for RTP socket to %s:%i.",rtp_addr,rtp_port);
 		return -1;
 	}
 	
 	memset(&hints, 0, sizeof(hints));
-	hints.ai_family = PF_UNSPEC;
+	hints.ai_family = (session->rtp.socket == -1) ? AF_UNSPEC : session->rtp.sockfamily;
 	hints.ai_socktype = SOCK_DGRAM;
+	hints.ai_flags = hints.ai_family==AF_INET6 ? AI_V4MAPPED : 0;
 	snprintf(num, sizeof(num), "%d", rtcp_port);
-	err = getaddrinfo(addr, num, &hints, &res0);
+	err = getaddrinfo(rtcp_addr, num, &hints, &res0);
 	if (err) {
 		ortp_warning ("Error: %s", gai_strerror(err));
 		return err;
@@ -695,14 +842,13 @@ rtp_session_set_remote_addr_full (RtpSession * session, const char * addr, int r
 	}
 	freeaddrinfo(res0);
 	if (err) {
-		ortp_warning("Could not set destination for RCTP socket to %s:%i.",addr,rtcp_port);
+		ortp_warning("Could not set destination for RCTP socket to %s:%i.",rtcp_addr,rtcp_port);
 		return -1;
 	}
 #else
 	session->rtp.rem_addrlen=sizeof(session->rtp.rem_addr);
 	session->rtp.rem_addr.sin_family = AF_INET;
-
-	err = inet_aton (addr, &session->rtp.rem_addr.sin_addr);
+	err = inet_aton (rtp_addr, &session->rtp.rem_addr.sin_addr);
 	if (err < 0)
 	{
 		ortp_warning ("Error in socket address:%s.", getSocketError());
@@ -710,10 +856,15 @@ rtp_session_set_remote_addr_full (RtpSession * session, const char * addr, int r
 	}
 	session->rtp.rem_addr.sin_port = htons (rtp_port);
 
-	memcpy (&session->rtcp.rem_addr, &session->rtp.rem_addr,
-		sizeof (struct sockaddr_in));
-	session->rtcp.rem_addr.sin_port = htons (rtcp_port);
 	session->rtcp.rem_addrlen=sizeof(session->rtcp.rem_addr);
+	session->rtcp.rem_addr.sin_family = AF_INET;
+	err = inet_aton (rtcp_addr, &session->rtcp.rem_addr.sin_addr);
+	if (err < 0)
+	{
+		ortp_warning ("Error in socket address:%s.", getSocketError());
+		return err;
+	}
+	session->rtcp.rem_addr.sin_port = htons (rtcp_port);
 #endif
 	if (can_connect(session)){
 		if (try_connect(session->rtp.socket,(struct sockaddr*)&session->rtp.rem_addr,session->rtp.rem_addrlen))
@@ -736,12 +887,16 @@ rtp_session_set_remote_addr_full (RtpSession * session, const char * addr, int r
 		session->flags&=~RTP_SOCKET_CONNECTED;
 		session->flags&=~RTCP_SOCKET_CONNECTED;
 	}
+	ortp_message("rtp session [%p] set to rtp [%s:%i] rtcp [%s:%i]"	,session
+																	,rtp_addr
+																	,rtp_port
+																	,rtcp_addr
+																	,rtcp_port);
 	return 0;
 }
 
-int
-rtp_session_set_remote_addr_and_port(RtpSession * session, const char * addr, int rtp_port, int rtcp_port){
-	return rtp_session_set_remote_addr_full(session,addr,rtp_port,rtcp_port);
+int rtp_session_set_remote_addr_and_port(RtpSession * session, const char * addr, int rtp_port, int rtcp_port){
+	return rtp_session_set_remote_addr_full(session,addr,rtp_port,addr,rtcp_port);
 }
 
 void rtp_session_set_sockets(RtpSession *session, int rtpfd, int rtcpfd)
@@ -839,26 +994,36 @@ static int rtp_sendmsg(int sock,mblk_t *m, struct sockaddr *rem_addr, int addr_l
 #define IP_UDP_OVERHEAD (20+8)
 #define IP6_UDP_OVERHEAD (40+8)
 
+#ifdef ORTP_INET6
+static bool_t is_ipv6(RtpSession *s){
+	if (s->rtp.sockfamily==AF_INET6){
+		struct sockaddr_in6 *in6=(struct sockaddr_in6*)&s->rtp.rem_addr;
+		return !IN6_IS_ADDR_V4MAPPED(&in6->sin6_addr);
+	}
+	return FALSE;
+}
+#endif
+
 static void update_sent_bytes(RtpSession*s, int nbytes){
 #ifdef ORTP_INET6
-	int overhead=(s->rtp.sockfamily==AF_INET6) ? IP6_UDP_OVERHEAD : IP_UDP_OVERHEAD;
+	int overhead=is_ipv6(s) ? IP6_UDP_OVERHEAD : IP_UDP_OVERHEAD;
 #else
 	int overhead=IP_UDP_OVERHEAD;
 #endif
 	if (s->rtp.sent_bytes==0){
-		gettimeofday(&s->rtp.send_bw_start,NULL);
+		ortp_gettimeofday(&s->rtp.send_bw_start,NULL);
 	}
 	s->rtp.sent_bytes+=nbytes+overhead;
 }
 
 static void update_recv_bytes(RtpSession*s, int nbytes){
 #ifdef ORTP_INET6
-	int overhead=(s->rtp.sockfamily==AF_INET6) ? IP6_UDP_OVERHEAD : IP_UDP_OVERHEAD;
+	int overhead=is_ipv6(s) ? IP6_UDP_OVERHEAD : IP_UDP_OVERHEAD;
 #else
 	int overhead=IP_UDP_OVERHEAD;
 #endif
 	if (s->rtp.recv_bytes==0){
-		gettimeofday(&s->rtp.recv_bw_start,NULL);
+		ortp_gettimeofday(&s->rtp.recv_bw_start,NULL);
 	}
 	s->rtp.recv_bytes+=nbytes+overhead;
 }
@@ -874,12 +1039,16 @@ rtp_session_rtp_send (RtpSession * session, mblk_t * m)
 	ortp_socket_t sockfd=session->rtp.socket;
 
 	hdr = (rtp_header_t *) m->b_rptr;
-	/* perform host to network conversions */
-	hdr->ssrc = htonl (hdr->ssrc);
-	hdr->timestamp = htonl (hdr->timestamp);
-	hdr->seq_number = htons (hdr->seq_number);
-	for (i = 0; i < hdr->cc; i++)
-		hdr->csrc[i] = htonl (hdr->csrc[i]);
+	if (hdr->version == 0) {
+		/* We are probably trying to send a STUN packet so don't change its content. */
+	} else {
+		/* perform host to network conversions */
+		hdr->ssrc = htonl (hdr->ssrc);
+		hdr->timestamp = htonl (hdr->timestamp);
+		hdr->seq_number = htons (hdr->seq_number);
+		for (i = 0; i < hdr->cc; i++)
+			hdr->csrc[i] = htonl (hdr->csrc[i]);
+	}
 
 	if (session->flags & RTP_SOCKET_CONNECTED) {
 		destaddr=NULL;
@@ -954,8 +1123,104 @@ rtp_session_rtcp_send (RtpSession * session, mblk_t * m)
 	return error;
 }
 
-int
-rtp_session_rtp_recv (RtpSession * session, uint32_t user_ts)
+int rtp_session_rtp_recv_abstract(ortp_socket_t socket, mblk_t *msg, int flags, struct sockaddr *from, socklen_t *fromlen) {
+	int ret;
+	int bufsz = (int) (msg->b_datap->db_lim - msg->b_datap->db_base);
+#ifndef _WIN32
+	struct iovec   iov;
+	struct msghdr  msghdr;
+	struct cmsghdr *cmsghdr;
+	struct {
+			struct cmsghdr cm;
+			char control[512];
+		} control;
+	memset(&msghdr, 0, sizeof(msghdr));
+	memset(&iov, 0, sizeof(iov));
+	iov.iov_base = msg->b_wptr;
+	iov.iov_len  = bufsz;
+	if(from != NULL && fromlen != NULL) {
+		msghdr.msg_name = from;
+		msghdr.msg_namelen = *fromlen;
+	}
+	msghdr.msg_iov     = &iov;
+	msghdr.msg_iovlen  = 1;
+	msghdr.msg_control = &control;
+	msghdr.msg_controllen = sizeof(control);
+
+	ret = recvmsg(socket, &msghdr, flags);
+	if(fromlen != NULL)
+		*fromlen = msghdr.msg_namelen;
+	if(ret >= 0) {
+#else
+	char control[512];
+	WSAMSG msghdr;
+	WSACMSGHDR *cmsghdr;
+	WSABUF data_buf;
+	DWORD bytes_received;
+
+	if (ortp_WSARecvMsg == NULL) {
+		return recvfrom(socket, (char *)msg->b_wptr, bufsz, flags, from, fromlen);
+	}
+
+	memset(&msghdr, 0, sizeof(msghdr));
+	memset(control, 0, sizeof(control));
+	if(from != NULL && fromlen != NULL) {
+		msghdr.name = from;
+		msghdr.namelen = *fromlen;
+	}
+	data_buf.buf = (char *)msg->b_wptr;
+	data_buf.len = bufsz;
+	msghdr.lpBuffers = &data_buf;
+	msghdr.dwBufferCount = 1;
+	msghdr.Control.buf = control;
+	msghdr.Control.len = sizeof(control);
+	msghdr.dwFlags = flags;
+	ret = ortp_WSARecvMsg(socket, &msghdr, &bytes_received, NULL, NULL);
+	if(fromlen != NULL)
+		*fromlen = msghdr.namelen;
+	if(ret >= 0) {
+		ret = bytes_received;
+#endif
+		for (cmsghdr = CMSG_FIRSTHDR(&msghdr); cmsghdr != NULL ; cmsghdr = CMSG_NXTHDR(&msghdr, cmsghdr)) {
+#if defined(ORTP_TIMESTAMP)
+			if (cmsghdr->cmsg_level == SOL_SOCKET && cmsghdr->cmsg_type == SO_TIMESTAMP) {
+				memcpy(&msg->timestamp, (struct timeval *)CMSG_DATA(cmsghdr), sizeof(struct timeval));
+			}
+#endif
+#ifdef IP_PKTINFO
+			if ((cmsghdr->cmsg_level == IPPROTO_IP) && (cmsghdr->cmsg_type == IP_PKTINFO)) {
+				struct in_pktinfo *pi = (struct in_pktinfo *)CMSG_DATA(cmsghdr);
+				memcpy(&msg->recv_addr.addr.ipi_addr, &pi->ipi_addr, sizeof(msg->recv_addr.addr.ipi_addr));
+				msg->recv_addr.family = AF_INET;
+			}
+#endif
+#ifdef IPV6_PKTINFO
+			if ((cmsghdr->cmsg_level == IPPROTO_IPV6) && (cmsghdr->cmsg_type == IPV6_PKTINFO)) {
+				struct in6_pktinfo *pi = (struct in6_pktinfo *)CMSG_DATA(cmsghdr);
+				memcpy(&msg->recv_addr.addr.ipi6_addr, &pi->ipi6_addr, sizeof(msg->recv_addr.addr.ipi6_addr));
+				msg->recv_addr.family = AF_INET6;
+			}
+#endif
+#ifdef IP_RECVDSTADDR
+			if ((cmsghdr->cmsg_level == IPPROTO_IP) && (cmsghdr->cmsg_type == IP_RECVDSTADDR)) {
+				struct in_addr *ia = (struct in_addr *)CMSG_DATA(cmsghdr);
+				memcpy(&msg->recv_addr.addr.ipi_addr, ia, sizeof(msg->recv_addr.addr.ipi_addr));
+				msg->recv_addr.family = AF_INET;
+			}
+#endif
+#ifdef IPV6_RECVDSTADDR
+			if ((cmsghdr->cmsg_level == IPPROTO_IPV6) && (cmsghdr->cmsg_type == IPV6_RECVDSTADDR)) {
+				struct in6_addr *ia = (struct in6_addr *)CMSG_DATA(cmsghdr);
+				memcpy(&msg->recv_addr.addr.ipi6_addr, ia, sizeof(msg->recv_addr.addr.ipi6_addr));
+				msg->recv_addr.family = AF_INET6;
+			}
+#endif
+		}
+	}
+	return ret;
+}
+
+int rtp_session_rtp_recv (RtpSession * session, uint32_t user_ts)
 {
 	int error;
 	ortp_socket_t sockfd=session->rtp.socket;
@@ -971,26 +1236,25 @@ rtp_session_rtp_recv (RtpSession * session, uint32_t user_ts)
 
 	while (1)
 	{
-		int bufsz;
 		bool_t sock_connected=!!(session->flags & RTP_SOCKET_CONNECTED);
 
 		if (session->rtp.cached_mp==NULL)
 			 session->rtp.cached_mp = msgb_allocator_alloc(&session->allocator,session->recv_buf_size);
 		mp=session->rtp.cached_mp;
-		bufsz=(int) (mp->b_datap->db_lim - mp->b_datap->db_base);
 		if (sock_connected){
-			error=recv(sockfd,(char*)mp->b_wptr,bufsz,0);
-		}else if (rtp_session_using_transport(session, rtp)) 
+			error=rtp_session_rtp_recv_abstract(sockfd, mp, 0, NULL, NULL);
+		}else if (rtp_session_using_transport(session, rtp)) {
 			error = (session->rtp.tr->t_recvfrom)(session->rtp.tr, mp, 0,
 				  (struct sockaddr *) &remaddr,
 				  &addrlen);
-		else error = recvfrom(sockfd, (char*)mp->b_wptr,
-				  bufsz, 0,
+		} else { error = rtp_session_rtp_recv_abstract(sockfd, mp, 0,
 				  (struct sockaddr *) &remaddr,
 				  &addrlen);
+		}
 		if (error > 0){
-			if (session->symmetric_rtp && !sock_connected){
-				if (session->use_connect){
+			if (session->use_connect){
+				/* In the case where use_connect is false, symmetric RTP is handled in rtp_session_rtp_parse() */
+				if (session->symmetric_rtp && !sock_connected){
 					/* store the sender rtp address to do symmetric RTP */
 					memcpy(&session->rtp.rem_addr,&remaddr,addrlen);
 					session->rtp.rem_addrlen=addrlen;
@@ -1003,8 +1267,8 @@ rtp_session_rtp_recv (RtpSession * session, uint32_t user_ts)
 				mp=rtp_session_network_simulate(session,mp);
 			/* then parse the message and put on jitter buffer queue */
 			if (mp){
-				rtp_session_rtp_parse(session, mp, user_ts, (struct sockaddr*)&remaddr,addrlen);
 				update_recv_bytes(session,mp->b_wptr-mp->b_rptr);
+				rtp_session_rtp_parse(session, mp, user_ts, (struct sockaddr*)&remaddr,addrlen);	
 			}
 			session->rtp.cached_mp=NULL;
 			/*for bandwidth measurements:*/
@@ -1017,6 +1281,13 @@ rtp_session_rtp_recv (RtpSession * session, uint32_t user_ts)
 				if (session->on_network_error.count>0){
 					rtp_signal_table_emit3(&session->on_network_error,(long)"Error receiving RTP packet",INT_TO_POINTER(getSocketErrorCode()));
 				}else ortp_warning("Error receiving RTP packet: %s, err num  [%i],error [%i]",getSocketError(),errnum,error);
+#ifdef __ios
+				/*hack for iOS and non-working socket because of background mode*/
+				if (errnum==ENOTCONN){
+					/*re-create new sockets */
+					rtp_session_set_local_addr(session,session->rtp.sockfamily==AF_INET ? "0.0.0.0" : "::0",session->rtp.loc_port,session->rtcp.loc_port);
+				}
+#endif
 			}else{
 				/*EWOULDBLOCK errors or transports returning 0 are ignored.*/
 				if (session->net_sim_ctx){
@@ -1024,8 +1295,8 @@ rtp_session_rtp_recv (RtpSession * session, uint32_t user_ts)
 					mp=rtp_session_network_simulate(session,NULL);
 					if (mp){
 						/* then parse the message and put on jitter buffer queue */
-						rtp_session_rtp_parse(session, mp, user_ts, (struct sockaddr*)&session->rtp.rem_addr,session->rtp.rem_addrlen);
 						update_recv_bytes(session,msgdsize(mp));
+						rtp_session_rtp_parse(session, mp, user_ts, (struct sockaddr*)&session->rtp.rem_addr,session->rtp.rem_addrlen);						
 					}
 				}
 			}
@@ -1051,17 +1322,19 @@ static void compute_rtt(RtpSession *session, const struct timeval *now, const re
 	uint32_t approx_ntp=(curntp>>16) & 0xFFFFFFFF;
 	uint32_t last_sr_time=report_block_get_last_SR_time(rb);
 	uint32_t sr_delay=report_block_get_last_SR_delay(rb);
-	/*ortp_message("rtt curntp=%u, last_sr_time=%u, sr_delay=%u",approx_ntp,last_sr_time,sr_delay);*/
+	/*ortp_message("rtt approx_ntp=%u, last_sr_time=%u, sr_delay=%u",approx_ntp,last_sr_time,sr_delay);*/
 	if (last_sr_time!=0 && sr_delay!=0){
 		double rtt_frac=approx_ntp-last_sr_time-sr_delay;
-		rtt_frac/=65536.0;
-		/*take into account the network simulator */
-		if (session->net_sim_ctx && session->net_sim_ctx->params.max_bandwidth>0){
-			double sim_delay=(double)session->net_sim_ctx->qsize/(double)session->net_sim_ctx->params.max_bandwidth;
-			rtt_frac+=sim_delay;
-		}
-		session->rtt=rtt_frac;
-		/*ortp_message("rtt estimated to %f ms",session->rtt);*/
+		if (rtt_frac>=0){
+			rtt_frac/=65536.0;
+			/*take into account the network simulator */
+			if (session->net_sim_ctx && session->net_sim_ctx->params.max_bandwidth>0){
+				double sim_delay=(double)session->net_sim_ctx->qsize/(double)session->net_sim_ctx->params.max_bandwidth;
+				rtt_frac+=sim_delay;
+			}
+			session->rtt=rtt_frac;
+			/*ortp_message("rtt estimated to %f s",session->rtt);*/
+		}else ortp_warning("Negative RTT computation, maybe due to clock adjustments.");
 	}
 }
 
@@ -1069,26 +1342,49 @@ static void compute_rtt(RtpSession *session, const struct timeval *now, const re
  * @brief : for SR packets, retrieves their timestamp, gets the date, and stores these information into the session descriptor. The date values may be used for setting some fields of the report block of the next RTCP packet to be sent.
  * @param session : the current session descriptor.
  * @param block : the block descriptor that may contain a SR RTCP message.
+ * @return -1 if we detect that the packet is in fact a STUN packet, otherwise 0.
  * @note a basic parsing is done on the block structure. However, if it fails, no error is returned, and the session descriptor is left as is, so it does not induce any change in the caller procedure behaviour.
  */
-static void process_rtcp_packet( RtpSession *session, mblk_t *block ) {
+static int process_rtcp_packet( RtpSession *session, mblk_t *block, struct sockaddr *addr, socklen_t addrlen ) {
 	rtcp_common_header_t *rtcp;
 	RtpStream * rtpstream = &session->rtp;
 
 	int msgsize = (int) ( block->b_wptr - block->b_rptr );
 	if ( msgsize < RTCP_COMMON_HEADER_SIZE ) {
 		ortp_debug( "Receiving a too short RTCP packet" );
-		return;
+		return 0;
 	}
 
 	rtcp = (rtcp_common_header_t *)block->b_rptr;
+
+	if (rtcp->version != 2)
+	{
+		/* try to see if it is a STUN packet */
+		uint16_t stunlen = *((uint16_t *)(block->b_rptr + sizeof(uint16_t)));
+		stunlen = ntohs(stunlen);
+		if (stunlen + 20 == block->b_wptr - block->b_rptr) {
+			/* this looks like a stun packet */
+			if (session->eventqs != NULL) {
+				OrtpEvent *ev = ortp_event_new(ORTP_EVENT_STUN_PACKET_RECEIVED);
+				OrtpEventData *ed = ortp_event_get_data(ev);
+				ed->packet = block;
+				ed->ep = rtp_endpoint_new(addr, addrlen);
+				ed->info.socket_type = OrtpRTCPSocket;
+				rtp_session_dispatch_event(session, ev);
+				return -1;
+			}
+		}
+		/* discard in two case: the packet is not stun OR nobody is interested by STUN (no eventqs) */
+		ortp_debug("Receiving rtcp packet with version number !=2...discarded");
+		return 0;
+	}
 	/* compound rtcp packet can be composed by more than one rtcp message */
 	do{
 		struct timeval reception_date;
 		const report_block_t *rb;
 		
 		/* Getting the reception date from the main clock */	
-		gettimeofday( &reception_date, NULL );
+		ortp_gettimeofday( &reception_date, NULL );
 
 		if (rtcp_is_SR(block) ) {
 			rtcp_sr_t *sr = (rtcp_sr_t *) rtcp;
@@ -1101,12 +1397,12 @@ static void process_rtcp_packet( RtpSession *session, mblk_t *block ) {
 			
 			if ( ntohl( sr->ssrc ) != session->rcv.ssrc ) {
 				ortp_debug( "Receiving a RTCP SR packet from an unknown ssrc" );
-				return;
+				return 0;
 			}
 
 			if ( msgsize < RTCP_COMMON_HEADER_SIZE + RTCP_SSRC_FIELD_SIZE + RTCP_SENDER_INFO_SIZE + ( RTCP_REPORT_BLOCK_SIZE * sr->ch.rc ) ) {
 				ortp_debug( "Receiving a too short RTCP SR packet" );
-				return;
+				return 0;
 			}
 
 			/* Saving the data to fill LSR and DLSR field in next RTCP report to be transmitted */
@@ -1123,6 +1419,7 @@ static void process_rtcp_packet( RtpSession *session, mblk_t *block ) {
 		}
 	}while (rtcp_next_packet(block));
 	rtcp_rewind(block);
+	return 0;
 }
 
 
@@ -1149,7 +1446,7 @@ rtp_session_rtcp_recv (RtpSession * session)
 		
 		mp=session->rtcp.cached_mp;
 		if (sock_connected){
-			error=recv(session->rtcp.socket,(char*)mp->b_wptr,RTCP_MAX_RECV_BUFSIZE,0);
+			error=rtp_session_rtp_recv_abstract(session->rtcp.socket, mp, 0, NULL, NULL);
 		}else {
 			addrlen=sizeof (remaddr);
 
@@ -1158,15 +1455,14 @@ rtp_session_rtcp_recv (RtpSession * session)
 				  (struct sockaddr *) &remaddr,
 				  &addrlen);
 			else
-			  error=recvfrom (session->rtcp.socket,(char*) mp->b_wptr,
-				  RTCP_MAX_RECV_BUFSIZE, 0,
+			  error=rtp_session_rtp_recv_abstract (session->rtcp.socket,mp, 0,
 				  (struct sockaddr *) &remaddr,
 				  &addrlen);
 		}
 		if (error > 0)
 		{
 			mp->b_wptr += error;
-			process_rtcp_packet( session, mp );
+			if (process_rtcp_packet( session, mp, (struct sockaddr*)&remaddr, addrlen) >= 0)
 			/* post an event to notify the application*/
 			{
 				rtp_session_notify_inc_rtcp(session,mp);
@@ -1196,7 +1492,7 @@ rtp_session_rtcp_recv (RtpSession * session)
 			{
 				if (session->on_network_error.count>0){
 					rtp_signal_table_emit3(&session->on_network_error,(long)"Error receiving RTCP packet",INT_TO_POINTER(errnum));
-				}else ortp_warning("Error receiving RTCP packet: %s.",getSocketError());
+				}else ortp_warning("Error receiving RTCP packet on port [%i]: %s.",session->rtcp.loc_port,getSocketError());
 				session->rtp.recv_errno=errnum;
 			}
 			/* don't free the cached_mp, it will be reused next time */
