@@ -17,16 +17,13 @@
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
-#include "ortp/port.h"
-#include "ortp/rtp.h"
 #include "ortp/str_utils.h"
-#include "utils.h"
+
 
 void qinit(queue_t *q){
-	mblk_init(&q->_q_first);
-	mblk_init(&q->_q_last);
-	q->_q_first.b_next=&q->_q_last;
-	q->_q_last.b_prev=&q->_q_first;
+	mblk_init(&q->_q_stopper);
+	q->_q_stopper.b_next=&q->_q_stopper;
+	q->_q_stopper.b_prev=&q->_q_stopper;
 	q->q_mcount=0;
 }
 
@@ -36,68 +33,89 @@ void mblk_init(mblk_t *mp)
 	mp->b_rptr=mp->b_wptr=NULL;
 	mp->reserved1=0;
 	mp->reserved2=0;
+#if defined(ORTP_TIMESTAMP)
+	memset(&(mp->timestamp), 0, sizeof(struct timeval));
+#endif
 }
 
-mblk_t *allocb(int size, int pri)
+void mblk_meta_copy(const mblk_t *source, mblk_t *dest) {
+	dest->reserved1 = source->reserved1;
+	dest->reserved2 = source->reserved2;
+#if defined(ORTP_TIMESTAMP)
+	dest->timestamp = source->timestamp;
+#endif
+}
+
+dblk_t *datab_alloc(size_t size){
+	dblk_t *db;
+	size_t total_size=sizeof(dblk_t)+size;
+	db=(dblk_t *) ortp_malloc(total_size);
+	db->db_base=(uint8_t*)db+sizeof(dblk_t);
+	db->db_lim=db->db_base+size;
+	db->db_ref=1;
+	db->db_freefn=NULL;	/* the buffer pointed by db_base must never be freed !*/
+	return db;
+}
+
+static inline void datab_ref(dblk_t *d){
+	d->db_ref++;
+}
+
+static inline void datab_unref(dblk_t *d){
+	d->db_ref--;
+	if (d->db_ref==0){
+		if (d->db_freefn!=NULL)
+			d->db_freefn(d->db_base);
+		ortp_free(d);
+	}
+}
+
+
+mblk_t *allocb(size_t size, int pri)
 {
 	mblk_t *mp;
 	dblk_t *datab;
-	char *buf;
-	
+
 	mp=(mblk_t *) ortp_malloc(sizeof(mblk_t));
 	mblk_init(mp);
-	datab=(dblk_t *) ortp_malloc(sizeof(dblk_t));
-	
-	buf=(char *) ortp_malloc(size);
+	datab=datab_alloc(size);
 
-	datab->db_base=buf;
-	datab->db_lim=buf+size;
-	datab->ref_count=1;
-	datab->db_freefn=ortp_free;
-	
 	mp->b_datap=datab;
-	mp->b_rptr=mp->b_wptr=buf;
+	mp->b_rptr=mp->b_wptr=datab->db_base;
 	mp->b_next=mp->b_prev=mp->b_cont=NULL;
 	return mp;
 }
 
-mblk_t *allocb_with_buf(char *buf, int size, int pri, void (*freefn)(void*) )
+mblk_t *esballoc(uint8_t *buf, size_t size, int pri, void (*freefn)(void*) )
 {
 	mblk_t *mp;
 	dblk_t *datab;
-	
+
 	mp=(mblk_t *) ortp_malloc(sizeof(mblk_t));
 	mblk_init(mp);
 	datab=(dblk_t *) ortp_malloc(sizeof(dblk_t));
-	
+
 
 	datab->db_base=buf;
 	datab->db_lim=buf+size;
-	datab->ref_count=1;
+	datab->db_ref=1;
 	datab->db_freefn=freefn;
-	
+
 	mp->b_datap=datab;
 	mp->b_rptr=mp->b_wptr=buf;
 	mp->b_next=mp->b_prev=mp->b_cont=NULL;
 	return mp;
 }
 
-	
+
 void freeb(mblk_t *mp)
 {
 	return_if_fail(mp->b_datap!=NULL);
 	return_if_fail(mp->b_datap->db_base!=NULL);
-	
-	mp->b_datap->ref_count--;
-	if (mp->b_datap->ref_count==0)
-	{
-		if (mp->b_datap->db_freefn!=NULL)
-			mp->b_datap->db_freefn(mp->b_datap->db_base);
-		
-		ortp_free(mp->b_datap);
-	}
+
+	datab_unref(mp->b_datap);
 	ortp_free(mp);
-}	
+}
 
 void freemsg(mblk_t *mp)
 {
@@ -116,10 +134,11 @@ mblk_t *dupb(mblk_t *mp)
 	mblk_t *newm;
 	return_val_if_fail(mp->b_datap!=NULL,NULL);
 	return_val_if_fail(mp->b_datap->db_base!=NULL,NULL);
-	
-	mp->b_datap->ref_count++;
+
+	datab_ref(mp->b_datap);
 	newm=(mblk_t *) ortp_malloc(sizeof(mblk_t));
 	mblk_init(newm);
+	mblk_meta_copy(mp, newm);
 	newm->b_datap=mp->b_datap;
 	newm->b_rptr=mp->b_rptr;
 	newm->b_wptr=mp->b_wptr;
@@ -143,20 +162,20 @@ mblk_t	*dupmsg(mblk_t* m)
 
 void putq(queue_t *q,mblk_t *mp)
 {
-	q->_q_last.b_prev->b_next=mp;
-	mp->b_prev=q->_q_last.b_prev;
-	mp->b_next=&q->_q_last;
-	q->_q_last.b_prev=mp;
+	q->_q_stopper.b_prev->b_next=mp;
+	mp->b_prev=q->_q_stopper.b_prev;
+	mp->b_next=&q->_q_stopper;
+	q->_q_stopper.b_prev=mp;
 	q->q_mcount++;
 }
 
 mblk_t *getq(queue_t *q)
 {
 	mblk_t *tmp;
-	tmp=q->_q_first.b_next;
-	if (tmp==&q->_q_last) return NULL;
-	q->_q_first.b_next=tmp->b_next;
-	tmp->b_next->b_prev=&q->_q_first;
+	tmp=q->_q_stopper.b_next;
+	if (tmp==&q->_q_stopper) return NULL;
+	q->_q_stopper.b_next=tmp->b_next;
+	tmp->b_next->b_prev=&q->_q_stopper;
 	tmp->b_prev=NULL;
 	tmp->b_next=NULL;
 	q->q_mcount--;
@@ -165,8 +184,8 @@ mblk_t *getq(queue_t *q)
 
 mblk_t * peekq(queue_t *q){
 	mblk_t *tmp;
-	tmp=q->_q_first.b_next;
-	if (tmp==&q->_q_last) return NULL;
+	tmp=q->_q_stopper.b_next;
+	if (tmp==&q->_q_stopper) return NULL;
 	return tmp;
 }
 
@@ -181,7 +200,7 @@ void insq(queue_t *q,mblk_t *emp, mblk_t *mp)
 	emp->b_prev->b_next=mp;
 	mp->b_prev=emp->b_prev;
 	emp->b_prev=mp;
-	mp->b_next=emp;	
+	mp->b_next=emp;
 }
 
 void remq(queue_t *q, mblk_t *mp){
@@ -196,56 +215,52 @@ void remq(queue_t *q, mblk_t *mp){
 void flushq(queue_t *q, int how)
 {
 	mblk_t *mp;
-	
+
 	while ((mp=getq(q))!=NULL)
 	{
 		freemsg(mp);
 	}
 }
 
-int msgdsize(mblk_t *mp)
+size_t msgdsize(const mblk_t *mp)
 {
-	int msgsize=0;
+	size_t msgsize=0;
 	while(mp!=NULL){
-		msgsize+=(int) (mp->b_wptr-mp->b_rptr);
+		msgsize+=(size_t) (mp->b_wptr-mp->b_rptr);
 		mp=mp->b_cont;
 	}
 	return msgsize;
 }
 
-mblk_t * msgpullup(mblk_t *mp,int len)
+void msgpullup(mblk_t *mp,size_t len)
 {
-	mblk_t *newm;
-	int msgsize=msgdsize(mp);
-	int rlen;
-	int mlen;
-	
-	
-	if ((len==-1) || (len>msgsize)) len=msgsize;
-	rlen=len;
-	newm=allocb(len,BPRI_MED);
+	mblk_t *firstm=mp;
+	dblk_t *db;
+	int wlen=0;
 
-	while(mp!=NULL){
-		mlen=(int) (mp->b_wptr-mp->b_rptr);
-		if (rlen>=mlen)
-		{
-			memcpy(newm->b_wptr,mp->b_rptr,mlen);
-			rlen-=mlen;
-			newm->b_wptr+=mlen;
+	if (mp->b_cont==NULL && len==-1) return;	/*nothing to do, message is not fragmented */
+
+	if (len==-1) len=msgdsize(mp);
+	db=datab_alloc(len);
+	while(wlen<len && mp!=NULL){
+		int remain=len-wlen;
+		int mlen=mp->b_wptr-mp->b_rptr;
+		if (mlen<=remain){
+			memcpy(&db->db_base[wlen],mp->b_rptr,mlen);
+			wlen+=mlen;
+			mp=mp->b_cont;
+		}else{
+			memcpy(&db->db_base[wlen],mp->b_rptr,remain);
+			wlen+=remain;
 		}
-		else /* rlen < mlen */
-		{
-			memcpy(newm->b_wptr,mp->b_rptr,rlen);
-			newm->b_wptr+=rlen;
-			
-			/* put the end of the original message at the end of the new */
-			newm->b_cont=dupmsg(mp);
-			newm->b_cont->b_rptr+=rlen;
-			return newm;
-		}
-		mp=mp->b_cont;
 	}
-	return newm;
+	/*set firstm to point to the new datab */
+	freemsg(firstm->b_cont);
+	firstm->b_cont=NULL;
+	datab_unref(firstm->b_datap);
+	firstm->b_datap=db;
+	firstm->b_rptr=db->db_base;
+	firstm->b_wptr=firstm->b_rptr+wlen;
 }
 
 
@@ -256,6 +271,7 @@ mblk_t *copyb(mblk_t *mp)
 	newm=allocb(len,BPRI_MED);
 	memcpy(newm->b_wptr,mp->b_rptr,len);
 	newm->b_wptr+=len;
+	memcpy(&newm->recv_addr,&mp->recv_addr,sizeof(newm->recv_addr));
 	return newm;
 }
 
@@ -272,15 +288,15 @@ mblk_t *copymsg(mblk_t *mp)
 	return newm;
 }
 
-mblk_t * appendb(mblk_t *mp, const char *data, int size, bool_t pad){
-	int padcnt=0;
-	int i;
+mblk_t * appendb(mblk_t *mp, const char *data, size_t size, bool_t pad){
+	size_t padcnt=0;
+	size_t i;
 	if (pad){
-		padcnt= (int)(4L-( (long)(((long)mp->b_wptr)+size) % 4L)) % 4L;
+		padcnt= (size_t)(4L-( (long)(((long)mp->b_wptr)+size) % 4L)) % 4L;
 	}
-	if ((mp->b_wptr + size +padcnt) > (char*)mp->b_datap->db_lim){
+	if ((mp->b_wptr + size +padcnt) > mp->b_datap->db_lim){
 		/* buffer is not large enough: append a new block (with the same size ?)*/
-		int plen=(int)((char*)mp->b_datap->db_lim - (char*) mp->b_datap->db_base);
+		size_t plen=(size_t)((char*)mp->b_datap->db_lim - (char*) mp->b_datap->db_base);
 		mp->b_cont=allocb(MAX(plen,size),0);
 		mp=mp->b_cont;
 	}
@@ -293,7 +309,7 @@ mblk_t * appendb(mblk_t *mp, const char *data, int size, bool_t pad){
 	return mp;
 }
 
-void msgappend(mblk_t *mp, const char *data, int size, bool_t pad){
+void msgappend(mblk_t *mp, const char *data, size_t size, bool_t pad){
 	while(mp->b_cont!=NULL) mp=mp->b_cont;
 	appendb(mp,data,size,pad);
 }
@@ -304,3 +320,29 @@ mblk_t *concatb(mblk_t *mp, mblk_t *newm){
 	while(newm->b_cont!=NULL) newm=newm->b_cont;
 	return newm;
 }
+
+void msgb_allocator_init(msgb_allocator_t *a){
+	qinit(&a->q);
+}
+
+mblk_t *msgb_allocator_alloc(msgb_allocator_t *a, size_t size){
+	queue_t *q=&a->q;
+	mblk_t *m,*found=NULL;
+
+	/*lookup for an unused msgb (data block with ref count ==1)*/
+	for(m=qbegin(q);!qend(q,m);m=qnext(q,m)){
+		if (m->b_datap->db_ref==1 && m->b_datap->db_lim-m->b_datap->db_base>=size){
+			found=m;
+			break;
+		}
+	}
+	if (found==NULL){
+		found=allocb(size,0);
+		putq(q,found);
+	}
+	return dupb(found);
+}
+
+void msgb_allocator_uninit(msgb_allocator_t *a){
+	flushq(&a->q,-1);
+}
